# ruff: noqa: F401
# This barrel file contains intentional import shortcuts
from .geometric_shapes import (
    Cone,
    Cuboid,
    Cylinder,
    Ellipsoid,
    GeometricShape,
    Parallelepiped,
    ShapesComposition,
    Sphere,
    inside_mbox,
)
from .morphology_shape_intersection import MorphologyToShapeIntersection
from .shape_morphology_intersection import ShapeToMorphologyIntersection
from .shape_shape_intersection import ShapeHemitype, ShapeToShapeIntersection



from .. import config
from .._util import obj_str_insert


@config.node
class SimulationComponent:
    name: str = config.attr(key=True)

    @property
    def simulation(self):
        return self._config_parent._config_parent

    @obj_str_insert
    def __str__(self):
        return f"'{self.name}'"


__all__ = ["SimulationComponent"]



import functools
import os

from ..exceptions import DependencyError
from ._util import MockModule


class MPIService:
    """
    Interface for MPI Communication context.

    This class will also emulate MPI Communication context in single node context.
    """

    def __init__(self, comm=None):
        self._mpi = MPIModule("mpi4py.MPI")
        self._comm = comm or self._mpi.COMM_WORLD

    def get_communicator(self):
        return self._comm

    def get_rank(self):
        if self._comm:
            return self._comm.Get_rank()
        return 0

    def get_size(self):
        if self._comm:
            return self._comm.Get_size()
        return 1

    def barrier(self):
        if self._comm:
            return self._comm.Barrier()
        pass

    def abort(self, errorcode=1):
        if self._comm:
            return self._comm.Abort(errorcode)
        print("MPI Abort called on MockCommunicator", flush=True)
        exit(errorcode)

    def bcast(self, obj, root=0):
        if self._comm:
            return self._comm.bcast(obj, root=root)
        return obj

    def gather(self, obj, root=0):
        if self._comm:
            return self._comm.gather(obj, root=root)
        return [obj]

    def allgather(self, obj):
        if self._comm:
            return self._comm.allgather(obj)
        return [obj]

    def window(self, buffer):
        if self._comm and self.get_size() > 1:
            from mpi4py.MPI import INFO_NULL, Win

            return Win.Create(buffer, True, INFO_NULL, self._comm)
        else:

            class WindowMock:
                def Get(self, bufspec, rank):
                    return bufspec[0]

                def Lock(self, rank):
                    pass

                def Unlock(self, rank):
                    pass

            return WindowMock()


class MPIModule(MockModule):
    """
    Module provider of the MPI interface.
    """

    @property
    @functools.cache
    def COMM_WORLD(self):
        if (
            any("mpi" in key.lower() for key in os.environ)
            and "BSB_IGNORE_PARALLEL" not in os.environ
        ):
            raise DependencyError(
                "MPI runtime detected without parallel support."
                + " Please install with `pip install bsb[parallel]`."
                + " Set `BSB_IGNORE_PARALLEL` to ignore this error."
            )
        return None



from itertools import chain

import numpy as np

from ... import config
from ...config import types
from ...exceptions import ConnectivityWarning
from ...reporting import warn
from ...storage._chunks import Chunk


class Intersectional:
    affinity = config.attr(type=types.fraction(), default=1)

    def get_region_of_interest(self, chunk):
        post_ps = [ct.get_placement_set() for ct in self.postsynaptic.cell_types]
        lpre, upre = self.presynaptic._get_rect_ext(tuple(chunk.dimensions))
        lpost, upost = self.postsynaptic._get_rect_ext(tuple(chunk.dimensions))
        # Get the `np.arange`s between bounds offset by the chunk position, to be used in
        # `np.meshgrid` below.
        bounds = list(
            np.arange(l1 - u2 + c, u1 - l2 + c + 1)
            for l1, l2, u1, u2, c in zip(lpre, lpost, upre, upost, chunk, strict=False)
        )
        # Flatten and stack the meshgrid coordinates into a list.
        clist = np.column_stack(
            [a.reshape(-1) for a in np.meshgrid(*bounds, indexing="ij")]
        )
        if not hasattr(self, "_occ_chunks"):
            # Filter by chunks where cells were actually placed
            self._occ_chunks = set(
                chain.from_iterable(ps.get_all_chunks() for ps in post_ps)
            )
        if not self._occ_chunks:
            warn(
                f"No {', '.join(ps.tag for ps in post_ps)} were placed, "
                f"skipping {self.name}",
                ConnectivityWarning,
            )
            return []
        else:
            size = next(iter(self._occ_chunks)).dimensions
            return [t for c in clist if (t := Chunk(c, size)) in self._occ_chunks]

    def candidate_intersection(self, target_coll, candidate_coll):
        target_cache = [
            (tset.cell_type, tset, tset.load_boxes()) for tset in target_coll.placement
        ]
        for cset in candidate_coll.placement:
            box_tree = cset.load_box_tree()
            for _ttype, tset, tboxes in target_cache:
                yield (tset, cset, self._affinity_filter(box_tree.query(tboxes)))

    def _affinity_filter(self, query):
        if self.affinity == 1:
            return query
        else:
            aff = self.affinity

            def sizemod(q):
                ln = len(q)
                return int(np.floor(ln * aff) + (np.random.rand() < ((ln * aff) % 1)))

            return (np.random.choice(q, sizemod(q), replace=False) for q in query)


__all__ = ["Intersectional"]



"""
Morphology module.
"""

# This is a note to myself, should expand into docs:
#
# It goes ``morphology-on-file`` into ``repository`` that the ``storage`` needs to provide
# support for. Then after a placement job has placed cells for a chunk, the positions are
# sent to a ``distributor`` that is supposed to use the ``indicators`` to ask the
# ``storage.morphology_repository`` which ``loaders`` are appropriate for the given
# ``selectors``, then, still hopefully using just morpho metadata the  ``distributor``
# generates indices and rotations. In more complex cases the ``selector`` and
# ``distributor`` can both load the morphologies but this will slow things down.
#
# In the simulation step, these (possibly dynamically modified) morphologies are passed
# to the cell model instantiators.

import functools
import inspect
import itertools
from collections import deque
from pathlib import Path
from pickle import UnpicklingError

import numpy as np
from scipy.spatial.transform import Rotation

from .. import _util as _gutil
from .._encoding import EncodedLabels
from ..exceptions import EmptyBranchError, MorphologyDataError, MorphologyError
from ..voxels import VoxelSet


class MorphologySet:
    """
    Associates a set of :class:`StoredMorphologies
    <bsb.storage.interfaces.StoredMorphology>` to cells.
    """

    def __init__(self, loaders, m_indices=None, /, labels=None):
        """
        :param loaders: list of Morphology loader functions.
        :type loaders: list[Callable[[], bsb.storage.interfaces.StoredMorphology]]
        :param m_indices: indices of the loaders for each of the morphologies.
        :type: list[int]
        """
        if m_indices is None:
            loaders, m_indices = np.unique(loaders, return_inverse=True)
        self._m_indices = np.array(m_indices, copy=False, dtype=int)
        self._loaders = list(loaders)
        check_max = np.max(m_indices, initial=-1)
        if check_max >= len(loaders):
            raise IndexError(f"Index {check_max} out of range for {len(loaders)}.")
        self._cached = {}
        self._labels = labels

    def set_label_filter(self, labels):
        self._cached = {}
        for loader in self._loaders:
            loader._cached_load.cache_clear()
        self._labels = labels

    @_gutil.obj_str_insert
    def __repr__(self):
        return f"{len(self)} cells, {len(self._loaders)} morphologies"

    def __contains__(self, value):
        return value in [loader.name for loader in self._loaders]

    def count_morphologies(self):
        return len(self._loaders)

    def count_unique(self):
        uniques = []
        count = 0
        for m in (m.load() for m in self._loaders):
            if not any(f == m for f in uniques):
                uniques.append(m)
                count += 1
        return count

    def __len__(self):
        return len(self._m_indices)

    def __iter__(self):
        return self.iter_morphologies()

    @property
    def names(self):
        return [loader.name for loader in self._loaders]

    def get_indices(self, copy=True):
        return self._m_indices.copy() if copy else self._m_indices

    def get(self, index, cache=True, hard_cache=False):
        data = self._m_indices[index]
        if data.ndim:
            return self._get_many(data, cache, hard_cache)
        else:
            return self._get_one(data, cache, hard_cache)

    def _get_one(self, idx, cache, hard_cache):
        if cache:
            if hard_cache:
                return self._loaders[idx].cached_load(self._labels)

            if idx not in self._cached:
                self._cached[idx] = (
                    self._loaders[idx].load().set_label_filter(self._labels).as_filtered()
                )
            return self._cached[idx].copy()
        else:
            return self._loaders[idx].load().set_label_filter(self._labels).as_filtered()

    def _get_many(self, data, cache, hard_cache):
        if hard_cache:
            return np.array([self._loaders[idx].cached_load() for idx in data])
        elif cache:
            res = []
            for idx in data:
                if idx not in self._cached:
                    self._cached[idx] = self._loaders[idx].load()
                res.append(self._cached[idx].copy())
        else:
            res = np.array([self._loaders[idx].load() for idx in data])
        return res

    def clear_soft_cache(self):
        self._cached = {}

    def iter_morphologies(self, cache=True, unique=False, hard_cache=False):
        """
        Iterate over the morphologies in a MorphologySet with full control over caching.

        :param cache: Use :ref:`Soft caching<soft-caching>` (1 copy stored in mem per
            cache miss, 1 copy created from that per cache hit).
        :type cache: bool
        :param hard_cache: Use :ref:`Hard caching<hard-caching>` (1 copy stored on the
            loader, always same copy returned from that loader forever).
        """
        if hard_cache:

            def _load(loader):
                return loader.cached_load(self._labels)

        elif self._labels is not None:

            def _load(loader):
                return loader.load().set_label_filter(self._labels).as_filtered()

        else:

            def _load(loader):
                return loader.load()

        if unique:
            yield from map(_load, self._loaders)
        elif not cache or hard_cache:
            yield from map(_load, (self._loaders[idx] for idx in self._m_indices))
        else:
            _cached = {}
            for idx in self._m_indices:
                if idx not in _cached:
                    _cached[idx] = _load(self._loaders[idx])
                yield _cached[idx].copy()

    def iter_meta(self, unique=False):
        if unique:
            yield from (loader.get_meta() for loader in self._loaders)
        else:
            yield from (self._loaders[idx].get_meta() for idx in self._m_indices)

    def _serialize_loaders(self):
        return [loader.name for loader in self._loaders]

    @classmethod
    def empty(cls):
        return cls([], np.empty(0, dtype=int))

    def merge(self, other):
        merged_loaders = self._loaders.copy()
        previous_set = set(merged_loaders)
        if any(loader in previous_set for loader in other._loaders):
            # There is overlap between the sets, and mapping is required
            id_map = dict(
                (i, merged_loaders.index(loader))
                for i, loader in enumerate(other._loaders)
                if loader in previous_set
            )
            if all(k == v for k, v in id_map.items()):
                mapped_indices = other._m_indices
            else:

                def map_ids(id):
                    mapped_id = id_map.get(id)
                    if mapped_id is None:
                        mapped_id = id_map[id] = len(merged_loaders)
                        merged_loaders.append(other._loaders[id])
                    return mapped_id

                mapped_indices = np.vectorize(map_ids)(other._m_indices)
            merged_indices = np.concatenate((self._m_indices, mapped_indices))
        else:
            # No overlap, we can just offset the new dataset
            merge_offset = len(self._loaders)
            merged_loaders = self._loaders + other._loaders
            merged_indices = np.concatenate(
                (self._m_indices, other._m_indices + merge_offset)
            )
        return MorphologySet(merged_loaders, merged_indices)

    def _mapback(self, locs):
        if self._labels is None:
            raise RuntimeError("Mapback requested on unfiltered morphology set.")
        locs = locs.copy()
        for i, loader in enumerate(self._loaders):
            rows = self._m_indices[locs[:, 0]] == i
            if np.any(rows):
                morpho = loader.load()
                filtered = morpho.set_label_filter(self._labels).as_filtered()
                # Using np.vectorize is Python speed O(n), worst case in numpy is C speed
                # O(n^2) (that is if every point is on another branch), not sure.
                branchmap = np.vectorize(
                    {
                        bid: b._copied_from_branch
                        for bid, b in enumerate(filtered.branches)
                    }.get
                )
                pointmap = np.vectorize(
                    {
                        bid: b._copied_points_offset
                        for bid, b in enumerate(filtered.branches)
                    }.get
                )
                # Map points first, then branches, since points depend on unmapped branch.
                locs[rows, 2] = locs[rows, 2] + pointmap(locs[rows, 1])
                locs[rows, 1] = branchmap(locs[rows, 1])
        return locs


class RotationSet:
    """
    Set of rotations.

    Returned rotations are of :class:`scipy.spatial.transform.Rotation`
    """

    def __init__(self, data):
        if not isinstance(data, np.ndarray):
            self._data = np.array(
                [
                    v.as_euler("xyz", degrees=True) if isinstance(v, Rotation) else v
                    for v in data
                ]
            )
        else:
            self._data = data
        if self._data.ndim != 2 or self._data.shape[1] != 3:
            raise ValueError("Input should be an (Nx3) matrix of rotations.")

    def __array__(self, dtype=None, *args, **kwargs):
        return self._data.__array__(dtype, *args, **kwargs)

    def __iter__(self):
        return self.iter()

    def __getitem__(self, index):
        data = self._data[index]
        if data.ndim == 2:
            return np.array([self._rot(d) for d in data])
        else:
            return self._rot(data)

    def __len__(self):
        return len(self._data)

    def iter(self, cache=False):
        if cache:
            yield from (self._cached_rot(tuple(d)) for d in self._data)
        else:
            yield from (self._rot(d) for d in self._data)

    @functools.cache
    def _cached_rot(self, angles):
        return self._rot(angles)

    def _rot(self, angles):
        return Rotation.from_euler("xyz", angles)


def branch_iter(branch):
    """
    Iterate over a branch and all of its children depth first.
    """
    yield branch
    for child in branch._children:
        yield from branch_iter(child)


class SubTree:
    """
    Collection of branches, not necessarily all connected.
    """

    def __init__(self, branches, sanitize=True):
        if sanitize:
            # Find the roots of the full subtree(s) emanating from the given, possibly
            # overlapping branches.
            if len(branches) < 2:
                # The roots of the subtrees of 0 or 1 branches is eaqual to the 0 or 1
                # branches themselves.
                self.roots = branches
            else:
                # Collect the deduplicated subtree emanating from all given branches; use
                # dict.fromkeys and .keys to preserve insertion order (i.e. DFS order)
                sub = dict.fromkeys(
                    itertools.chain(*(b.get_branches() for b in branches))
                ).keys()
                # Find the root branches whose parents are not part of the subtrees
                self.roots = [b for b in sub if b.parent not in sub]
        else:
            # No subtree sanitizing: Assume the whole tree is given, or only the roots
            # have been given, and just take all literal root (non-parent-having)
            # branches.
            self.roots = [b for b in branches if b.parent is None]
        self._is_shared = False

    def __getattr__(self, attr):
        if not hasattr(self, "_is_shared"):
            raise UnpicklingError("Morphology class does not support pickling.")
        if self._is_shared:
            if attr in self._shared._prop:
                return self._shared._prop[attr]
            else:
                super().__getattribute__(attr)
        else:
            return np.concatenate([getattr(b, attr) for b in self.branches])

    def __len__(self):
        if self._is_shared:
            return len(self._shared._points)
        else:
            return sum(b.size for b in self.get_branches())

    @property
    def size(self):
        return len(self)

    @property
    def branches(self):
        """
        Return a depth-first flattened array of all branches.
        """
        return self.get_branches()

    @property
    def points(self):
        return self.flatten()

    @points.setter
    def points(self, value):
        arr = np.array(value, copy=False, dtype=float)
        if self._is_shared:
            self.points[:] = arr
        else:
            ptr = 0
            for b in self.branches:
                b.points = arr[ptr : (ptr := ptr + len(b))]

    @property
    def radii(self):
        return self.flatten_radii()

    @radii.setter
    def radii(self, value):
        arr = np.array(value, copy=False, dtype=float)
        if self._is_shared:
            self.radii[:] = arr
        else:
            ptr = 0
            for b in self.branches:
                b.radii = arr[ptr : (ptr := ptr + len(b))]

    @property
    def labels(self):
        return self.flatten_labels()

    @property
    def properties(self):
        return self.flatten_properties()

    @property
    def bounds(self):
        f = self.flatten()
        if not len(f):
            return np.zeros(3), np.zeros(3)
        return np.min(f, axis=0), np.max(f, axis=0)

    @property
    def branch_adjacency(self):
        """
        Return a dictionary containing mapping the id of the branch to its children.
        """
        idmap = {b: n for n, b in enumerate(self.branches)}
        return {n: list(map(idmap.get, b.children)) for n, b in enumerate(self.branches)}

    @property
    def path_length(self):
        """
        Return the total path length as the sum of the euclidian distances between
        consecutive points.
        """
        return sum(b.path_length for b in self.branches)

    def subtree(self, labels=None):
        return SubTree(self.get_branches(labels))

    def get_branches(self, labels=None):
        """
        Return a depth-first flattened array of all or the selected branches.

        :param list labels: Names of the labels to select.
        :returns: List of all branches, or the ones fully labelled with any of the given
            labels.
        :rtype: list
        """
        root_iter = (branch_iter(root) for root in self.roots)
        all_branch = itertools.chain(*root_iter)
        if labels is None:
            return [*all_branch]
        else:
            return [b for b in all_branch if b.contains_labels(labels)]

    def flatten(self):
        """
        Return the flattened points of the morphology or subtree.

        :rtype: numpy.ndarray
        """
        if self._is_shared:
            return self._shared._points
        branches = self.get_branches()
        if not branches:
            return np.empty((0, 3))
        return np.vstack(tuple(b.points for b in branches))

    def flatten_radii(self):
        """
        Return the flattened radii of the morphology or subtree.

        :rtype: numpy.ndarray
        """
        if self._is_shared:
            return self._shared._radii
        branches = self.get_branches()
        if not branches:
            return np.empty(0)
        return np.concatenate(tuple(b.radii for b in branches))

    def flatten_labels(self):
        """
        Return the flattened labels of the morphology or subtree.

        :rtype: numpy.ndarray
        """
        if self._is_shared:
            return self._shared._labels
        else:
            return EncodedLabels.concatenate(*(b._labels for b in self.get_branches()))

    def flatten_properties(self):
        """
        Return the flattened properties of the morphology or subtree.

        :rtype: numpy.ndarray
        """
        if self._is_shared:
            if not self._shared._prop:
                return {}
            return self._shared._prop.copy()
        else:
            branches = self.get_branches()
            len_ = sum(len(b) for b in branches)
            all_props = [*set(_gutil.ichain(b._properties.keys() for b in branches))]
            props = {k: np.empty(len_) for k in all_props}
            ptr = 0
            for branch in self.branches:
                nptr = ptr + len(branch)
                for k, v in props.items():
                    prop = branch._properties.get(k, None)
                    if prop is None:
                        prop = np.full(len(branch), np.nan)
                    v[ptr:nptr] = prop
                ptr = nptr
            return props

    def label(self, labels, points=None):
        """
        Add labels to the morphology or subtree.

        :param labels: Labels to add to the subtree.
        :type labels: list[str]
        :param points: Optional boolean or integer mask for the points to be labelled.
        :type points: numpy.ndarray
        """
        if points is None:
            points = np.ones(len(self), dtype=bool)
        points = np.array(points, copy=False)
        if self._is_shared:
            self.labels.label(labels, points)
        else:
            if len(points) == len(self) and points.dtype == bool:
                ctr = 0
                for b in self.branches:
                    b.label(labels, points[ctr : ctr + len(b)])
                    ctr += len(b)
            elif np.can_cast(points.dtype, int):
                points = np.array(points, copy=False, dtype=int)
                for b in self.branches:
                    mux = points < len(b)
                    b.label(labels, points[mux])
                    points = points[~mux]
            else:
                raise ValueError(
                    "Label indices must be a boolean or integer mask."
                    f" {points.dtype} given."
                )

        return self

    def rotate(self, rotation, center=None):
        """
        Rotate the entire Subtree with respect to the center. The rotation angles are
        assumed to be in degrees. If the center is not provided, the Subtree will rotate
        from [0, 0, 0].

        :param rotation: Scipy rotation
        :type rotation: scipy.spatial.transform.Rotation | list[float,float,float]
        :param center: rotation offset point.
        :type center: numpy.ndarray
        """
        if not isinstance(rotation, Rotation):
            rotation = Rotation.from_euler("xyz", rotation, degrees=True)
        if self._is_shared:
            self._shared._points[:] = self._rotate(self._shared._points, rotation, center)
        else:
            for b in self.branches:
                b.points[:] = self._rotate(b.points, rotation, center)
        return self

    def _rotate(self, points, rot, center):
        if center is not None:
            points = points - center
            rotated_points = rot.apply(points)
            rotated_points = rotated_points + center
        else:
            rotated_points = rot.apply(points)
        return rotated_points

    def root_rotate(self, rot, downstream_of=0):
        """
        Rotate the subtree emanating from each root around the start of that root If
        downstream_of is provided, will rotate points starting from the index provided
        (only for subtrees with a single root).

        :param rot: Scipy rotation to apply to the subtree.
        :type rot: scipy.spatial.transform.Rotation
        :param downstream_of: index of the point in the subtree from which the rotation
            should be applied. This feature works only when the subtree has only one root
            branch.
        :returns: rotated Morphology
        :rtype: bsb.morphologies.SubTree
        """

        if downstream_of != 0:
            if len(self.roots) > 1:
                raise ValueError(
                    "Can't rotate with subbranch precision with multiple roots"
                )
            elif type(downstream_of) is int and 0 < downstream_of < len(
                self.roots[0].points
            ):
                b = self.roots[0]
                group = SubTree([b])
                upstream = np.copy(b.points[:downstream_of])
                group.rotate(rot, b.points[downstream_of])
                b.points[:downstream_of] = upstream
        else:
            for b in self.roots:
                group = SubTree([b])
                group.rotate(rot, group.origin)
        return self

    def translate(self, point):
        """
        Translate the subtree by a 3D vector.

        :param numpy.ndarray point: 3D vector to translate the subtree.
        :returns: the translated subtree
        :rtype: bsb.morphologies.SubTree
        """
        if len(point) != 3:
            raise ValueError("Point must be a sequence of x, y and z coordinates")
        if self._is_shared:
            self._shared._points[:] = self._shared._points + point
        else:
            for branch in self.branches:
                branch.points[:] = branch.points[:] + point
        return self

    @property
    def origin(self):
        return np.mean([r.points[0] for r in self.roots], axis=0)

    def center(self):
        """
        Center the morphology on the origin.
        """
        self.translate(-self.origin)
        return self

    def close_gaps(self):
        """
        Close any head-to-tail gaps between parent and child branches.
        """
        for branch in self.branches:
            if branch.parent is not None:
                gap_offset = branch.parent.points[-1] - branch.points[0]
                if not np.allclose(gap_offset, 0):
                    SubTree([branch]).translate(gap_offset)
        return self

    def collapse(self, on=None):
        """
        Collapse all the roots of the morphology or subtree onto a single point.

        :param on: Index of the root to collapse on. Collapses onto the origin by default.
        :type on: int
        """
        if on is None:
            on = self.origin
        for root in self.roots:
            root.translate(on - root.points[0])
        return self

    def simplify_branches(self, epsilon):
        """
        Apply Ramer–Douglas–Peucker algorithm to all points of all branches of the
        SubTree.

        :param epsilon: Epsilon to be used in the algorithm.
        """
        for branch in self.branches:
            branch.simplify(epsilon)

    def voxelize(self, N):
        """
        Turn the morphology or subtree into an approximating set of axis-aligned cuboids.

        :rtype: bsb.voxels.VoxelSet
        """
        return VoxelSet.from_morphology(self, N)

    @functools.cache
    def cached_voxelize(self, N):
        """
        Turn the morphology or subtree into an approximating set of axis-aligned cuboids
        and cache the result.

        :rtype: bsb.voxels.VoxelSet
        """
        return self.voxelize(N)


class _SharedBuffers:
    def __init__(self, points, radii, labels, properties):
        self._points = points
        self._radii = radii
        self._labels = labels if labels is not None else EncodedLabels.none(len(radii))
        self._prop = properties

    def copy(self):
        copied_props = {k: v.copy() for k, v in self._prop.items()}
        return self.__class__(
            self._points.copy(), self._radii.copy(), self._labels.copy(), copied_props
        )

    def points_shared(self, branches):
        return all(b.points.base is self._points for b in branches)

    def radii_shared(self, branches):
        return all(b.radii.base is self._radii for b in branches)

    def labels_shared(self, branches):
        return all(b.labels.base is self._labels for b in branches)

    def properties_shared(self, branches):
        return all(
            (
                b._properties.keys() == self._prop.keys() and all(c.base is self._prop[c])
                for a, c in b._properties.items()
            )
            for b in branches
        )

    def all_buffers_shared(self, branches):
        return (
            self.points_shared(branches)
            and self.radii_shared(branches)
            and self.labels_shared(branches)
            and self.properties_shared(branches)
        )

    def get_shared(self, start, end):
        copied_props = {k: v[start:end] for k, v in self._prop.items()}
        return (
            self._points[start:end],
            self._radii[start:end],
            self._labels[start:end],
            copied_props,
        )


class Morphology(SubTree):
    """
    A multicompartmental spatial representation of a cell based on a directed acyclic
    graph of branches whom consist of data vectors, each element of a vector being a
    coordinate or other associated data of a point on the branch.
    """

    def __init__(self, roots, meta=None, shared_buffers=None, sanitize=False):
        super().__init__(roots, sanitize=sanitize)
        self._meta = meta if meta is not None else {}
        self._filter = None
        if shared_buffers is None:
            self._shared = None
            self._is_shared = False
        else:
            if isinstance(shared_buffers, _SharedBuffers):
                self._shared = shared_buffers
            else:
                self._shared = _SharedBuffers(*shared_buffers)
            self._is_shared = self._check_shared()
            for branch in self.branches:
                branch._on_mutate = self._mutnotif

    @_gutil.obj_str_insert
    def __repr__(self):
        return (
            f"{len(self.roots)} roots, {len(self)} points,"
            f" from {self.bounds[0]} to {self.bounds[1]}"
        )

    def __eq__(self, other):
        return len(self.branches) == len(other.branches) and all(
            b1.is_terminal == b2.is_terminal and (not b1.is_terminal or b1 == b2)
            for b1, b2 in zip(self.branches, other.branches, strict=False)
        )

    def __lt__(self, other):
        # Sorting compares using lt, so we use id for useless but stable comparison.
        return id(self) < id(other)

    def __hash__(self):
        return id(self)

    def _check_shared(self):
        if self._shared is None:
            return False
        return self._shared.all_buffers_shared(self.branches)

    def _mutnotif(self):
        self._is_shared = False

    @property
    def is_optimized(self):
        return self._shared

    def optimize(self, force=False):
        if force or not self._is_shared:
            branches = self.branches
            len_ = sum(len(b) for b in branches)
            points = np.empty((len_, 3))
            radii = np.empty(len_)
            all_props = [*set(_gutil.ichain(b._properties.keys() for b in branches))]
            types = [
                next(_p[k].dtype for b in branches if k in (_p := b._properties))
                for k in all_props
            ]
            props = {
                k: np.empty(len_, dtype=t) for k, t in zip(all_props, types, strict=False)
            }
            labels = EncodedLabels.concatenate(*(b._labels for b in branches))
            ptr = 0
            for branch in self.branches:
                nptr = ptr + len(branch)
                points[ptr:nptr] = branch.points
                branch._points = points[ptr:nptr]
                radii[ptr:nptr] = branch.radii
                branch._radii = radii[ptr:nptr]
                for k, v in props.items():
                    prop = branch._properties.get(k, None)
                    if prop is None:
                        prop = np.full(len(branch), np.nan)
                    v[ptr:nptr] = prop
                    branch._properties[k] = v[ptr:nptr]
                branch._labels = labels[ptr:nptr]
                ptr = nptr
            self._shared = _SharedBuffers(points, radii, labels, props)
            self._is_shared = True
            assert self._check_shared(), "optimize should result in shared buffers"

    @property
    def meta(self):
        return self._meta

    @meta.setter
    def meta(self, value):
        self._meta = value

    @property
    def adjacency_dictionary(self):
        """
        Return a dictonary associating to each key (branch index) a list of adjacent
        branch indices.
        """
        branches = self.branches
        idmap = {b: n for n, b in enumerate(branches)}
        return {n: list(map(idmap.get, b.children)) for n, b in enumerate(branches)}

    @property
    def labelsets(self):
        """
        Return the sets of labels associated to each numerical label.
        """
        self.optimize()
        return self._shared._labels.labels

    def list_labels(self):
        """
        Return a list of labels present on the morphology.
        """
        self.optimize()
        return sorted(set(_gutil.ichain(self._shared._labels.labels.values())))

    def set_label_filter(self, labels):
        """
        Set a label filter, so that `as_filtered` returns copies filtered by these labels.
        """
        self._filter = labels
        return self

    def get_label_mask(self, labels):
        """
        Get a mask corresponding to all the points labelled with 1 or more of the given
        labels.
        """
        self.optimize()
        return self.labels.get_mask(labels)

    @classmethod
    def empty(cls):
        return cls([])

    def copy(self):
        """
        Copy the morphology.
        """
        # Make sure to optimize so that we use 1 shared buffer.
        self.optimize(force=False)
        # Copy that buffer into a new one
        buffers = self._shared.copy()
        roots = []
        branch_copy_map = {}
        ptr = 0
        # For each branch, create a copy, and assign a piece of the copied buffer to it.
        # Also attach it to its intended parent. Since we iterate DFS, each parent occurs
        # before their children, so we can always find it in `branch_copy_map` from a
        # previous iteration.
        for branch in self.branches:
            nptr = ptr + len(branch)
            nbranch = Branch(*buffers.get_shared(ptr, nptr))
            branch_copy_map[branch] = nbranch
            if not branch.is_root:
                branch_copy_map[branch.parent].attach_child(nbranch)
            else:
                roots.append(nbranch)
            ptr = nptr
        # Construct the morphology
        return self.__class__(roots, shared_buffers=buffers, meta=self.meta.copy())

    def as_filtered(self, labels=None):
        """
        Return a filtered copy of the morphology that includes only points that match the
        current label filter, or the specified labels.
        """
        filter = labels if labels is not None else self._filter
        if filter is None:
            return self.copy()
        self.optimize(force=False)
        buffers = self._shared.copy()
        roots = []
        branch_copy_map = {None: None}
        ptr = 0
        # Iterate over each branch, and turn it into 0 or more branches.
        for og_id, branch in enumerate(self.branches):
            # Using the filter mask, figure out where to split the branch into pieces.
            # Parts, or the entire branch, may be excluded if there are no labelled points
            filtered = branch.get_label_mask(filter)
            # Each boolean in filtered represents a point, either included or excluded.
            # Every subbranch begins where a point is excluded and the next point is
            # included, and ends where a point is included, and the next point is excluded
            starts = (np.nonzero(filtered[1:] & ~filtered[:-1])[0] + 1).tolist()
            ends = (np.nonzero(filtered[:-1] & ~filtered[1:])[0] + 1).tolist()
            # Treat the boundary.
            if len(filtered) and filtered[0]:
                starts.insert(0, 0)
            if len(filtered) and filtered[-1]:
                ends.append(len(filtered))
            prev = None
            nbranch = None
            # Make all the sub branches. Connect the first to the parent, and store the
            # last in the map, for children to be connected to.
            for start, end in zip(starts, ends, strict=False):
                nbranch = Branch(*buffers.get_shared(ptr + start, ptr + end))
                # Store where this branch came from, for loc mapping.
                nbranch._copied_from_branch = og_id
                # Store where the points map to
                nbranch._copied_points_offset = start
                if not prev:
                    if branch.is_root or branch_copy_map[branch.parent] is None:
                        roots.append(nbranch)
                    else:
                        branch_copy_map[branch.parent].attach_child(nbranch)
                else:
                    prev.attach_child(nbranch)
                prev = nbranch
            ptr = ptr + len(branch)
            if nbranch is None:
                # If an entire branch is unlabelled, skip it, and map our children's
                # parent to their grandparent, since we, the parent, don't exist.
                branch_copy_map[branch] = branch_copy_map[branch.parent]
            else:
                # Did we create some branches? Use the last iteration value as parent for
                # our children.
                branch_copy_map[branch] = nbranch
        # Construct and return the morphology
        return self.__class__(roots, meta=self.meta.copy())

    def swap_axes(self, axis1: int, axis2: int):
        """
        Interchange two axes of a morphology points.

        :param int axis1: index of the first axis to exchange
        :param int axis2: index of the second axis to exchange
        :return: the modified morphology
        :rtype: bsb.morphologies.Morphology
        """
        if not 0 <= axis1 < 3 or not 0 <= axis2 < 3:
            raise ValueError(
                f"Axes values should be in [0, 1, 2], {axis1}, {axis2} given."
            )
        for b in self.branches:
            old_column = np.copy(b.points[:, axis1])
            b.points[:, axis1] = b.points[:, axis2]
            b.points[:, axis2] = old_column

        return self

    def simplify(self, *args, optimize=True, **kwargs):
        super().simplify_branches(*args, **kwargs)
        if optimize:
            self.optimize()

    def to_swc(self, file):
        """
        Create a SWC file from a Morphology.

        :param file: path to write to
        """
        file_data = _morpho_to_swc(self)
        if isinstance(file, str | Path):
            np.savetxt(
                file,
                file_data,
                fmt="%d %d %f %f %f %f %d",
                delimiter="\t",
                newline="\n",
                header="",
                footer="",
                comments="# ",
                encoding=None,
            )

    def to_graph_array(self):
        """
        Create a SWC-like numpy array from a Morphology.

        .. warning::

            Custom SWC tags (above 3) won't work and throw an error

        :returns: a numpy array with columns storing the standard SWC attributes
        :rtype: numpy.ndarray
        """
        data = _morpho_to_swc(self)
        return data


def _copy_api(cls, wrap=lambda self: self):
    # Wraps functions, so they are called with `self` wrapped in `wrap`
    def make_wrapper(f):
        @functools.wraps(f)
        def wrapper(self, *args, **kwargs):
            return f(wrap(self), *args, **kwargs)

        return wrapper

    # Decorates a class so that it copies and wraps (see above) the public API of `cls`
    def decorator(decorated_cls):
        for key, f in vars(cls).items():
            if (
                inspect.isfunction(f)
                and not key.startswith("_")
                and key not in vars(decorated_cls)
            ):
                setattr(decorated_cls, key, make_wrapper(f))

        return decorated_cls

    return decorator


# For every `SubTree.f` there is a `Branch.f` == `SubTree([branch]).f` so we copy and wrap
# the public API of `SubTree` onto `Branch`, with the `SubTree([self])` wrapped into it.
@_copy_api(SubTree, lambda self: SubTree([self]))
class Branch:
    """
    A vector based representation of a series of point in space.

    Can be a root or connected to a parent branch. Can be a terminal branch or have
    multiple children.
    """

    def __init__(self, points, radii, labels=None, properties=None, children=None):
        """
        :param points: Array of 3D coordinates defining the point of the branch
        :type points: list | numpy.ndarray
        :param radii: Array of radii associated to each point
        :type radii: list | numpy.ndarray
        :param labels: Array of labels to associate to each point
        :type labels: list[str] | set | numpy.ndarray
        :param properties: dictionary of per-point data to store in the branch
        :type properties: dict
        :param children: list of child branches to attach to the branch
        :type children: list[bsb.morphologies.Branch]
        :raises bsb.exceptions.MorphologyError: if a property of the branch does not have
            the same size as its points
        """

        self._points = _gutil.sanitize_ndarray(points, (-1, 3), float)
        self._radii = _gutil.sanitize_ndarray(radii, (-1,), float)
        _gutil.assert_samelen(self._points, self._radii)
        self._children = []
        if labels is None:
            labels = EncodedLabels.none(len(points))
        elif not isinstance(labels, EncodedLabels):
            labels = EncodedLabels.from_labelset(len(points), labels)
        self._labels = labels
        if properties is None:
            properties = {}
        mismatched = [str(k) for k, v in properties.items() if len(v) != len(points)]
        if mismatched:
            raise MorphologyError(
                f"Morphology properties {', '.join(mismatched)} are not "
                f"length {len(points)}"
            )
        self._properties = {
            k: v if isinstance(v, np.ndarray) else np.array(v)
            for k, v in properties.items()
        }
        self._parent = None
        self._on_mutate = lambda: None
        if children is not None:
            for child in children:
                self.attach_child(child)

    def set_properties(self, **kwargs):
        for prop, values in kwargs.items():
            if len(values) != len(self):
                raise ValueError(f"Expected {len(self)} {prop}, got {len(values)}.")
            if prop in self._properties:
                self._properties[prop][:] = values
            else:
                self._on_mutate()
                self._properties[prop] = values

    def __getattr__(self, attr):
        if not hasattr(self, "_properties"):
            raise UnpicklingError("Branch class does not support pickling.")
        if attr in self._properties:
            return self._properties[attr]
        else:
            super().__getattribute__(attr)

    def __copy__(self):
        return self.copy()

    def __bool__(self):
        # Without this, empty branches are False, and `if branch.parent:` checks fail.
        return True

    def __eq__(self, other):
        if isinstance(other, Branch):
            return (
                self.points.shape == other.points.shape
                and np.allclose(self.points, other.points)
                and self.labels == other.labels
            )
        else:
            return np.allclose(self.points, other)

    def __hash__(self):
        return id(self)

    @property
    def parent(self):
        return self._parent

    @property
    def size(self):
        """
        Returns the amount of points on this branch.

        :returns: Number of points on the branch.
        :rtype: int
        """
        return len(self._points)

    def __len__(self):
        return self.size

    @property
    def points(self):
        """
        Return the spatial coordinates of the points on this branch.
        """
        return self._points

    @points.setter
    def points(self, value):
        arr = np.array(value, copy=False, dtype=float)
        if arr.shape == self._points.shape:
            self._points[:] = arr
        elif arr.ndim != 2 or arr.shape[1] != 3:
            raise ValueError(f"Point data must have (N, 3) shape, {arr.shape} given.")
        else:
            self._points = arr

    @property
    def _kd_tree(self):
        """
        Return a `scipy.spatial.cKDTree` of this branch points for fast spatial queries.

        .. warning::

           Constructing a kd-tree takes time and should only be used for repeat queries.
        """
        import scipy.spatial

        return scipy.spatial.cKDTree(self._points)

    @property
    def point_vectors(self):
        """
        Return the individual vectors between consecutive points on this branch.
        """
        return np.diff(self.points, axis=0)

    @property
    def segments(self):
        """
        Return the start and end points of vectors between consecutive points on this
        branch.
        """
        return np.hstack(
            (self.points[:-1], self.points[:-1] + self.point_vectors)
        ).reshape(-1, 2, 3)

    @property
    def start(self):
        """
        Return the spatial coordinates of the starting point of this branch.
        """
        try:
            return self._points[0]
        except IndexError:
            raise EmptyBranchError("Empty branch has no starting point") from None

    @property
    def end(self):
        """
        Return the spatial coordinates of the terminal point of this branch.
        """
        try:
            return self._points[-1]
        except IndexError:
            raise EmptyBranchError("Empty branch has no ending point") from None

    @property
    def vector(self):
        """
        Return the vector of the axis connecting the start and terminal points.
        """
        try:
            return self.end - self.start
        except IndexError:
            raise EmptyBranchError("Empty branch has no vector") from None

    @property
    def versor(self):
        """
        Return the normalized vector of the axis connecting the start and terminal points.
        """
        if np.all(self.end == self.start):
            raise EmptyBranchError("Empty and single-point branched have no versor")
        else:
            return (self.end - self.start) / np.linalg.norm(self.end - self.start)

    @property
    def euclidean_dist(self):
        """
        Return the Euclidean distance from the start to the terminal point of this branch.
        """
        try:
            return np.sqrt(np.sum((self.end - self.start) ** 2))
        except IndexError:
            raise EmptyBranchError("Empty branch has no Euclidean distance") from None

    @property
    def max_displacement(self):
        """
        Return the max displacement of the branch points from its axis vector.
        """
        try:
            displacements = np.linalg.norm(
                np.cross(self.versor, (self.points - self.start)), axis=1
            )
            return np.max(displacements)
        except IndexError:
            raise EmptyBranchError(
                "Impossible to compute max_displacement in branches with 0 or 1 points."
            ) from None

    @property
    def path_length(self):
        """
        Return the sum of the euclidean distances between the points on the branch.
        """
        return np.sum(np.sqrt(np.sum(self.point_vectors**2, axis=1)))

    @property
    def fractal_dim(self):
        """
        Return the fractal dimension of this branch, computed as the coefficient of the
        line fitting the log-log plot of path vs euclidean distances of its points.
        """
        if len(self.points) == 0:
            raise EmptyBranchError("Empty branch has no fractal dimension") from None
        else:
            euclidean = np.sqrt(np.sum((self.points - self.start) ** 2, axis=1))
            path = np.cumsum(np.sqrt(np.sum(self.point_vectors**2, axis=1)))
            log_e = np.log(euclidean[1:])
            log_p = np.log(path)
            if len(self.points) <= 2:
                return 1.0
            return np.polyfit(log_e, log_p, 1)[0]

    @property
    def radii(self):
        """
        Return the radii of the points on this branch.
        """
        return self._radii

    @radii.setter
    def radii(self, value):
        arr = np.array(value, copy=False, dtype=float)
        if arr.shape == self._radii.shape:
            self._radii[:] = arr
        else:
            self._radii = arr.ravel()

    @property
    def labels(self):
        """
        Return the labels of the points on this branch.

        Labels are represented as a number that is associated to a set of labels. See
        :ref:`Labels <morphology_labels>` for more info.
        """
        return self._labels

    @property
    def labelsets(self):
        """
        Return the sets of labels associated to each numerical label.
        """
        return self._labels.labels

    def list_labels(self):
        """
        Return a list of labels present on the branch.
        """
        lookup = np.vectorize(self._labels.labels.get)
        labels = np.unique(lookup(self._labels.raw))
        return sorted(set(_gutil.ichain(labels)))

    @property
    def is_root(self):
        """
        Returns whether this branch is root or if it has a parent.

        :returns: True if this branch has no parent, False otherwise.
        :rtype: bool
        """
        return not self._parent

    @property
    def is_terminal(self):
        """
        Returns whether this branch is terminal or if it has children.

        :returns: True if this branch has no children, False otherwise.
        :rtype: bool
        """
        return not self._children

    def copy(self, branch_class=None):
        """
        Return a parentless and childless copy of the branch.

        :param branch_class: Custom branch creation class
        :type branch_class: type
        :returns: A branch, or `branch_class` if given, without parents or children.
        :rtype: bsb.morphologies.Branch
        """
        cls = branch_class or type(self)
        props = {k: v.copy() for k, v in self._properties.items()}
        return cls(self._points.copy(), self._radii.copy(), self._labels.copy(), props)

    def label(self, labels, points=None):
        """
        Add labels to the branch.

        :param labels: Label(s) for the branch
        :type labels: list[str]
        :param points: An integer or boolean mask to select the points to label.
        """
        if points is None:
            points = np.ones(len(self), dtype=bool)
        self._labels.label(labels, points)

    @property
    def children(self):
        """
        Collection of the child branches of this branch.

        :returns: list of :class:`Branches <bsb.morphologies.Branch>`
        :rtype: list
        """
        return self._children.copy()

    def attach_child(self, branch):
        """
        Attach a branch as a child to this branch.

        :param branch: Child branch
        :type branch: :class:`Branch <bsb.morphologies.Branch>`
        """
        self._on_mutate()
        if branch._parent is not None:
            branch._parent.detach_child(branch)
        self._children.append(branch)
        branch._parent = self

    def find_closest_point(self, coord):
        """
        Return the index of the closest on this branch to a desired coordinate.

        :param coord: The coordinate to find the nearest point to
        :type: :class:`numpy.ndarray`
        """
        diff = np.sqrt(np.sum((self._points - coord) ** 2, axis=1))
        return np.argmin(diff)

    def insert_branch(self, branch, index):
        """
        Split this branch and insert the given ``branch`` at the specified ``index``.

        :param branch: Branch to be attached
        :type branch: :class:`Branch <bsb.morphologies.Branch>`
        :param index: Index or coordinates of the cutpoint; if coordinates are given, the
            closest point to the coordinates is used.
        :type: numpy.ndarray | int
        """
        index = np.array(index, copy=False)
        if index.ndim != 0:
            index = self.find_closest_point(index)

        if index < 0 or index >= len(self):
            raise IndexError(
                f"Cannot insert branch at cutpoint: "
                f"index {index} is out of range ({len(self)})"
            )

        if index == len(self.points) - 1:
            self.attach_child(branch)
        elif index == 0:
            self.parent.attach_child(branch)
        else:
            first_segment = Branch(
                self._points.copy()[: index + 1],
                self._radii.copy()[: index + 1],
                self._labels.copy()[: index + 1],
                {k: v.copy()[: index + 1] for k, v in self._properties.items()},
            )
            self.parent.attach_child(first_segment)
            self.parent.detach_child(self)
            first_segment.attach_child(branch)
            second_segment = Branch(
                self._points.copy()[index:],
                self._radii.copy()[index:],
                self._labels.copy()[index:],
                {k: v.copy()[index:] for k, v in self._properties.items()},
            )
            for b in self.children:
                self.detach_child(b)
                second_segment.attach_child(b)
            first_segment.attach_child(second_segment)

    def detach(self):
        """
        Detach the branch from its parent, if one exists.
        """
        if self.parent:
            self.parent.detach_child(self)

    def detach_child(self, branch):
        """
        Remove a branch as a child from this branch.

        :param branch: Child branch
        :type branch: :class:`Branch <bsb.morphologies.Branch>`
        """
        if branch._parent is not self:
            raise ValueError(f"Can't detach {branch} from {self}, not a child branch.")
        self._on_mutate()
        self._children = [b for b in self._children if b is not branch]
        branch._parent = None

    def walk(self):
        """
        Iterate over the points in the branch.
        """
        return zip(
            self.points[:, 0],
            self.points[:, 1],
            self.points[:, 2],
            self.radii,
            self.labels.walk(),
            *self._properties.values(),
            strict=False,
        )

    def contains_labels(self, labels):
        """
        Check if this branch contains any points labelled with any of the given labels.

        :param labels: The labels to check for.
        :type labels: list[str]
        :rtype: bool
        """
        return self.labels.contains(labels)

    def get_points_labelled(self, labels):
        """
        Filter out all points with certain labels.

        :param labels: The labels to check for.
        :type labels: list[str] | numpy.ndarray[str]
        :returns: All points with the labels.
        :rtype: list[numpy.ndarray]
        """
        return self.points[self.get_label_mask(labels)]

    def get_label_mask(self, labels):
        """
        Return a mask for the specified labels.

        :param labels: The labels to check for.
        :type labels: list[str] | numpy.ndarray[str]
        :returns: A boolean mask that selects out the points that match the label.
        :rtype: list[numpy.ndarray]
        """
        return self.labels.get_mask(labels)

    def introduce_point(self, index, position, radius=None, labels=None, properties=None):
        """
        Insert a new point at ``index``, before the existing point at ``index``. Radius,
        labels and extra properties can be set or will be copied from the existing point
        at ``index``.

        :param index: Index of the new point.
        :type index: int
        :param position: Coordinates of the new point
        :type position: list[float]
        :param radius: The radius to assign to the point.
        :type radius: float
        :param labels: The labels to assign to the point.
        :type labels: list
        :param properties: The properties to assign to the point.
        :type properties: dict
        """
        if index < 0 or index >= len(self.points):
            raise IndexError(
                f"Could not introduce point in branch at index {index}: "
                f"out of bounds for branch length {len(self)}."
            )
        self._on_mutate()
        old_labels = self.labels[index]
        self.points = np.insert(self.points, index, position, 0)
        self._labels = np.insert(self._labels, index, old_labels)
        self._radii = np.insert(self._radii, index, radius or self._radii[index])
        # By default, duplicate the existing property value ...
        for k, v in self._properties.items():
            self._properties[k] = np.insert(v, index, v[index])
        if labels is not None:
            self.label(labels, [index])
        # ... and overwrite it with any new property values, if given.
        if properties is not None:
            for k, v in properties.items():
                if k in self._properties:
                    self._properties[k][index] = v
                else:
                    raise MorphologyError(
                        f"Property key '{k}' is not part of the Branch."
                    )

    def introduce_arc_point(self, arc_val):
        """
        Introduce a new point at the given arc length.

        :param arc_val: Arc length between 0 and 1 to introduce new point at.
        :type arc_val: float
        :returns: The index of the new point.
        :rtype: int
        """
        arc = self.as_arc()
        arc_point_floor = self.floor_arc_point(arc_val)
        arc_point_ceil = self.ceil_arc_point(arc_val)
        arc_floor = arc[arc_point_floor]
        arc_ceil = arc[arc_point_ceil]
        point_floor = self[arc_point_floor]
        point_ceil = self[arc_point_ceil]
        rem = (arc_val - arc_floor) / (arc_ceil - arc_floor)
        new_point = (point_ceil - point_floor) * rem + point_floor
        new_index = arc_point_floor + 1
        self.introduce_point(new_index, *new_point)
        return new_index

    def get_arc_point(self, arc, eps=1e-10):
        """
        Strict search for an arc point within an epsilon.

        :param arc: Arclength position to look for.
        :type arc: float
        :param eps: Maximum distance/tolerance to accept an arc point as a match.
        :type eps: float
        :returns: The matched arc point index, or ``None`` if no match is found
        :rtype: int | None
        """
        arc_values = self.as_arc()
        arc_match = (i for i, arc_p in enumerate(arc_values) if abs(arc_p - arc) < eps)
        return next(arc_match, None)

    def as_arc(self):
        """
        Return the branch as a vector of arclengths in the closed interval [0, 1]. An
        arclength is the distance each point to the start of the branch along the branch
        axis, normalized by total branch length. A point at the start will have an
        arclength close to 0, and a point near the end an arclength close to 1.

        :returns: Vector of branch points as arclengths.
        :rtype: :class:`numpy.ndarray`
        """
        arc_distances = np.sqrt(np.sum(np.diff(self.points, axis=0) ** 2, axis=1))
        arc_length = np.sum(arc_distances)
        return np.cumsum(np.concatenate(([0], arc_distances))) / arc_length

    def floor_arc_point(self, arc):
        """
        Get the index of the nearest proximal arc point.
        """
        p = 0
        for i, a in enumerate(self.as_arc()):
            if a <= arc:
                p = i
            else:
                break
        return p

    def ceil_arc_point(self, arc):
        """
        Get the index of the nearest distal arc point.
        """
        for i, a in enumerate(self.as_arc()):
            if a >= arc:
                return i
        return len(self) - 1

    def get_axial_distances(self, idx_start=0, idx_end=-1, return_max=False):
        """
        Return the displacements or its max value of a subset of branch points from its
        axis vector.

        :param int idx_start: index of the first point of the subset.
        :param int idx_end: index of the last point of the subset.
        :param bool return_max: if True the function only returns the max value of
            displacements, otherwise the entire array.
        """
        start = self.points[idx_start]
        end = self.points[idx_end]
        versor = (
            (end - start) / np.linalg.norm(end - start)
            if np.any(end != start)
            else np.zeros(3)
        )
        displacements = np.linalg.norm(
            np.cross(
                versor,
                (self.points[idx_start : idx_end + 1] - self.points[idx_start]),
            ),
            axis=1,
        )
        if return_max:
            try:
                return np.max(displacements)
            except IndexError:
                raise EmptyBranchError("Selected an empty subset of points") from None
        else:
            return displacements

    def delete_point(self, index):
        """
        Remove a point from the branch.

        :param int index: index position of the point to remove
        :returns: the branch where the point has been removed
        :rtype: bsb.morphologies.Branch
        """
        self._points = np.delete(self._points, index, axis=0)
        self._labels = np.delete(self._labels, index, axis=0)
        self._radii = np.delete(self._radii, index, axis=0)
        for k, v in self._properties.items():
            self._properties[k] = np.delete(v, index, axis=0)
        return self

    def simplify(self, epsilon, idx_start=0, idx_end=-1):
        """
        Apply Ramer–Douglas–Peucker algorithm to all points or a subset of points of the
        branch.

        :param epsilon: Epsilon to be used in the algorithm. :param idx_start = 0: Index
            of the first element of the subset of points to be reduced. :param epsilon =
            -1: Index of the last element of the subset of points to be reduced.
        """
        if len(self.points) < 3:
            return
        if idx_end == -1:
            idx_end = len(self.points) - 1
        if epsilon < 0:
            raise ValueError("Epsilon must be >= 0")

        reduced = []
        skipped = deque()

        while True:
            dists = self.get_axial_distances(idx_start, idx_end)
            try:
                idx_max = np.argmax(dists)
                dmax = dists[idx_max]
                idx_max = idx_start + idx_max
            except ValueError:
                dmax = 0

            reduced.append(idx_start)
            reduced.append(idx_end)
            if dmax > epsilon and len(dists) > 2:
                skipped.append((idx_max, idx_end))
                idx_end = idx_max - 1
            else:
                try:
                    idx_start, idx_end = skipped.pop()
                except IndexError:
                    break

        # sorted because indexes are appended to reduced from the middle of the list
        # (the first point with dist > epsilon)
        # then all points with smaller index  until 0, then all points with bigger index
        reduced = np.sort(np.unique(reduced))
        self.points = self.points[reduced]
        self.radii = self.radii[reduced]

    @functools.wraps(SubTree.cached_voxelize)
    @functools.cache
    def cached_voxelize(self, *args, **kwargs):
        return SubTree([self]).voxelize(*args, **kwargs)


def _morpho_to_swc(morpho):
    # Initialize an empty data array
    data = np.empty((len(morpho.points), 7), dtype=object)
    swc_tags = {"soma": 1, "axon": 2, "dendrites": 3}
    bmap = {}
    nid = 0
    offset = 0
    # Convert labels to tags
    if not hasattr(morpho, "tags"):
        tags = np.full(len(morpho.points), -1, dtype=int)
        for key in swc_tags:
            mask = morpho.get_label_mask([key])
            tags[mask] = swc_tags[key]
    else:
        tags = morpho.tags
    if np.any(tags == -1):
        raise NotImplementedError("Can't store morphologies with custom SWC tags")
    # Iterate over the morphology branches
    for offset, b in enumerate(morpho.branches):
        ids = (
            np.arange(nid, nid + len(b) - 1)
            if len(b) > 1
            else np.arange(nid, nid + len(b))
        )
        samples = ids + 1
        data[ids, 0] = samples
        data[ids, 1] = tags[ids + offset]
        data[ids, 2:5] = morpho.points[ids + offset]
        try:
            data[ids, 5] = morpho.radii[ids + offset]
        except Exception as e:
            raise MorphologyDataError(
                f"Couldn't convert morphology radii to SWC: {e}."
                " Note that SWC files cannot store multi-dimensional radii"
            ) from None
        nid += len(b) - 1 if len(b) > 1 else len(b)
        bmap[b] = ids[-1]
        data[ids, 6] = ids
        data[ids[0], 6] = -1 if b.parent is None else bmap[b.parent] + 1

    return data[data != np.array(None)].reshape(-1, 7)


__all__ = [
    "Branch",
    "Morphology",
    "MorphologySet",
    "RotationSet",
    "SubTree",
    "branch_iter",
]



from .. import config
from .component import SimulationComponent


@config.node
class ConnectionModel(SimulationComponent):
    tag: str = config.attr(type=str, key=True)

    def get_connectivity_set(self):
        return self.scaffold.get_connectivity_set(self.tag)


__all__ = ["ConnectionModel"]



import typing

import numpy as np

from .. import config
from ..config import types
from ..exceptions import ConnectivityError
from ..mixins import InvertedRoI
from .strategy import ConnectionStrategy

if typing.TYPE_CHECKING:
    from ..config import Distribution


@config.node
class Convergence(ConnectionStrategy):
    """
    Connect cells based on a convergence distribution, i.e. by connecting each source cell
    to X target cells.
    """

    convergence: "Distribution" = config.attr(type=types.distribution(), required=True)

    def connect(self):
        raise NotImplementedError("Needs to be restored, please open an issue.")


@config.node
class AllToAll(ConnectionStrategy):
    """
    All to all connectivity between two neural populations.
    """

    affinity: float = config.attr(type=types.float(0.0, 1.0), required=False, default=1.0)
    """
    Probability for each individual connection to be, default is 1, i.e. all connected.
    """

    def connect(self, pre, post):
        for from_ps in pre.placement:
            fl = len(from_ps)
            for to_ps in post.placement:
                len_ = len(to_ps)
                ml = fl * len_
                filtered_ = np.random.binomial(1, p=self.affinity, size=ml) > 0
                ml = np.count_nonzero(filtered_)
                src_locs = np.full((ml, 3), -1)
                dest_locs = np.full((ml, 3), -1)
                src_locs[:, 0] = np.repeat(np.arange(fl), len_)[filtered_]
                dest_locs[:, 0] = np.tile(np.arange(len_), fl)[filtered_]
                self.connect_cells(from_ps, to_ps, src_locs, dest_locs)


def _connect_fixed_degree(self, pre, post, degree, is_in):
    # Generalized connect function for Fixed in- and out-degree
    rng = np.random.default_rng()
    ps_counted = pre.placement if is_in else post.placement
    ps_fixed = post.placement if is_in else pre.placement
    high = sum(len(ps) for ps in ps_counted)
    if high < degree:
        raise ConnectivityError(
            f"Number of cells for dependant population ({high}) is too small to match "
            f"required degree value {degree} for connection strategy {self.name}"
        )
    for ps in ps_fixed:
        l_ = len(ps)
        counted_targets = np.full((l_ * degree, 3), -1)
        fixed_targets = np.full((l_ * degree, 3), -1)
        ptr = 0
        for i in range(l_):
            fixed_targets[ptr : ptr + degree, 0] = i
            counted_targets[ptr : ptr + degree, 0] = rng.choice(
                high, degree, replace=False
            )
            ptr += degree
        lowmux = 0
        for ps_o in ps_counted:
            highmux = lowmux + len(ps_o)
            demux_idx = (counted_targets[:, 0] >= lowmux) & (
                counted_targets[:, 0] < highmux
            )
            demuxed = counted_targets[demux_idx]
            demuxed[:, 0] -= lowmux
            if is_in:
                self.connect_cells(ps_o, ps, demuxed, fixed_targets[demux_idx])
            else:
                self.connect_cells(ps, ps_o, fixed_targets[demux_idx], demuxed)
            lowmux = highmux


@config.node
class FixedIndegree(InvertedRoI, ConnectionStrategy):
    """
    Connect a group of postsynaptic cell types to ``indegree`` uniformly random
    presynaptic cells from all the presynaptic cell types.
    """

    indegree: int = config.attr(type=int, required=True)
    """
    Number of postsynaptic cell to connect to each presynaptic cell.
    """

    def connect(self, pre, post):
        _connect_fixed_degree(self, pre, post, self.indegree, True)


@config.node
class FixedOutdegree(ConnectionStrategy):
    """
    Connect a group of presynaptic cell types to ``outdegree`` uniformly random
    postsynaptic cells from all the postsynaptic cell types.
    """

    outdegree: int = config.attr(type=int, required=True)
    """
    Number of presynaptic cell to connect to each postsynaptic cell.
    """

    def connect(self, pre, post):
        _connect_fixed_degree(self, pre, post, self.outdegree, False)


__all__ = ["AllToAll", "Convergence", "FixedIndegree", "FixedOutdegree"]



"""
An attrs-inspired class annotation system, but my A stands for amateuristic.
"""

import builtins
import contextlib
from functools import wraps

import errr

from ..exceptions import (
    BootError,
    CastError,
    CfgReferenceError,
    NoReferenceAttributeSignal,
    RequirementError,
)
from ._compile import _wrap_reserved
from ._hooks import run_hook
from ._make import (
    MISSING,
    _resolve_references,
    compile_class,
    compile_isc,
    compile_new,
    compile_postnew,
    make_copyable,
    make_dictable,
    make_get_node_name,
    make_tree,
    walk_nodes,
    wrap_root_postnew,
)


def root(root_cls):
    """
    Decorate a class as a configuration root node.
    """
    root_cls.attr_name = root_cls.node_name = r"{root}"
    root_cls._config_isroot = True
    return node(root_cls, root=True)


def node(node_cls, root=False, dynamic=False, pluggable=False):
    """
    Decorate a class as a configuration node.
    """
    # Recreate the class to set its metaclass a posteriori
    node_cls = compile_class(node_cls)
    node_cls._config_unset = []
    # Inherit the parent's attributes, if any exist on the class already
    attrs = getattr(node_cls, "_config_attrs", {}).copy()
    for k, v in builtins.dict(node_cls.__dict__).items():
        # Add our attributes
        if isinstance(v, ConfigurationAttribute):
            if v.unset:
                attrs.pop(k, None)
                delattr(node_cls, k)
                # Keep track of what this class wants to unset, in case of MRO traversal.
                node_cls._config_unset.append(k)
            else:
                attrs[k] = v
    node_cls._config_attrs = attrs
    node_cls.__post_new__ = compile_postnew(node_cls)
    node_cls._config_isroot = root
    if root:
        node_cls.__post_new__ = wrap_root_postnew(node_cls.__post_new__)
        node_cls._config_isbooted = False
    node_cls.__new__ = compile_new(
        node_cls, dynamic=dynamic, pluggable=pluggable, root=root
    )
    if dynamic:
        node_cls.__init_subclass__ = compile_isc(node_cls, dynamic)
    make_get_node_name(node_cls, root=root)
    make_tree(node_cls)
    make_dictable(node_cls)
    make_copyable(node_cls)

    return node_cls


def dynamic(
    node_cls=None,
    attr_name="cls",
    classmap=None,
    auto_classmap=False,
    classmap_entry=None,
    **kwargs,
):
    """
    Decorate a class to be castable to a dynamically configurable class using a class
    configuration attribute.

    *Example*: Register a required string attribute ``class`` (this is the default):

    .. code-block:: python

        @dynamic
        class Example:
            pass

    *Example*: Register a string attribute ``type`` with a default value
    'pkg.DefaultClass' as dynamic attribute:

    .. code-block:: python

        @dynamic(attr_name="type", required=False, default="pkg.DefaultClass")
        class Example:
            pass

    :param attr_name: Name under which to register the class attribute in the node.
    :type attr_name: str
    :param kwargs: All keyword arguments are passed to the constructor of the
      :func:`attribute <.config.attr>`.
    """
    if "required" not in kwargs and "default" not in kwargs:
        kwargs["required"] = True
    if "type" not in kwargs:
        kwargs["type"] = str
    class_attr = ConfigurationAttribute(**kwargs)
    dynamic_config = DynamicNodeConfiguration(classmap, auto_classmap, classmap_entry)
    if node_cls is None:
        # If node_cls is None, it means that no positional argument was given, which most
        # likely means that the @dynamic(...) syntax was used instead of the @dynamic.
        # This means we have to return an inner decorator instead of the decorated class
        def decorator(node_cls):
            return _dynamic(node_cls, class_attr, attr_name, dynamic_config)

        return decorator
    # Regular @dynamic syntax used, return decorated class
    return _dynamic(node_cls, class_attr, attr_name, dynamic_config)


class DynamicNodeConfiguration:
    def __init__(self, classmap=None, auto_classmap=False, entry=None):
        self.classmap = classmap
        self.auto_classmap = auto_classmap
        self.entry = entry


def _dynamic(node_cls, class_attr, attr_name, config):
    # Set the dynamic attribute
    setattr(node_cls, attr_name, class_attr)
    node_cls._config_dynamic_attr = attr_name
    # Other than that compile the dynamic class like a regular node class
    node_cls = node(node_cls, dynamic=config)

    if config.auto_classmap or config.classmap:
        node_cls._config_dynamic_classmap = config.classmap or {}
    # This adds the parent class to its own classmap, which for subclasses happens in init
    # subclass
    if config.entry is not None:
        if not hasattr(node_cls, "_config_dynamic_classmap"):
            raise ValueError(
                f"Calling `@config.dynamic` with `entry='{config.entry}'`"
                + " requires `classmap` or `auto_classmap` to be set as well"
                + f" on '{node_cls.__name__}'."
            )
        node_cls._config_dynamic_classmap[config.entry] = node_cls
    # Mark the class as its own dynamic root, (grand)child classes will all need to
    # inherit from this as an interface contract.
    node_cls._config_dynamic_root = node_cls
    return node_cls


def pluggable(key, plugin_name=None):
    """
    Create a node whose configuration is defined by a plugin.

    *Example*: If you want to use the :guilabel:`attr` to chose from all the installed
    `dbbs_scaffold.my_plugin` plugins:

    .. code-block:: python

        @pluggable("attr", "my_plugin")
        class PluginNode:
            pass

    This will then read :guilabel:`attr`, load the plugin and configure the node from
    the node class specified by the plugin.

    :param plugin_name: The name of the category of the plugin endpoint
    :type plugin_name: str
    """

    def inner_decorator(node_cls):
        node_cls._config_plugin_name = plugin_name
        node_cls._config_plugin_key = key
        class_attr = ConfigurationAttribute(type=str, required=True)
        setattr(node_cls, key, class_attr)
        return node(node_cls, pluggable=True)

    return inner_decorator


def attr(**kwargs):
    """
    Create a configuration attribute.

    Only works when used inside a class decorated with the :func:`node <.config.node>`,
    :func:`dynamic <.config.dynamic>`,  :func:`root <.config.root>` or  :func:`pluggable
    <.config.pluggable>` decorators.

    :param type: Type of the attribute's value.
    :type type: Callable
    :param required: Should an error be thrown if the attribute is not present?
    :type required: bool
    :param default: Default value.
    :type default: Any
    :param call_default: Should the default value be used (False) or called (True). Use
        this to prevent mutable default values.
    :type call_default: bool
    :param key: If set, the key of the parent is stored on this attribute.
    """
    return ConfigurationAttribute(**kwargs)


def ref(reference, **kwargs):
    """
    Create a configuration reference.

    Configuration references are attributes that transform their value into the value
    of another node or value in the document::

      {"keys": {"a": 3, "b": 5}, "simple_ref": "a"}

    With ``simple_ref = config.ref(lambda root, here: here["keys"])`` the value ``a``
    will be looked up in the configuration object (after all values have been cast) at
    the location specified by the callable first argument.
    """
    return ConfigurationReferenceAttribute(reference, **kwargs)


def reflist(reference, **kwargs):
    """
    Create a configuration reference list.
    """
    if "default" not in kwargs:
        kwargs["default"] = builtins.list
        kwargs["call_default"] = True
    return ConfigurationReferenceListAttribute(reference, **kwargs)


def slot(**kwargs):
    """
    Create an attribute slot that is required to be overriden by child or plugin classes.
    """
    return ConfigurationAttributeSlot(**kwargs)


def property(val=None, /, type=None, **kwargs):
    """
    Create a configuration property attribute.

    You may provide a value or a callable. Call
    `setter` on the return value as you would with a regular property.
    """
    if type is None:

        def type(v):
            return v

    def decorator(val):
        prop = val if callable(val) else lambda s: val
        return ConfigurationProperty(prop, type=type, **kwargs)

    if val is None:
        return decorator
    else:
        return decorator(val)


def provide(value):
    """
    Provide a value for a parent class' attribute.

    Can be a value or a callable, a readonly configuration property will be created from
    it either way.
    """
    prop = property(value)

    def provided(self, instance, value):
        raise AttributeError(f"Can't set attribute, class provides the value '{value}'.")

    # Create a callable object that invokes `provided` when called, and whose `bool()`
    # returns `False`. Later in `_is_settable_attr`, we use this to trick the short
    # circuiting logic, so that this setter doesn't make the internal logic set and error
    # out on this attr.
    prop.setter(
        type("provision", (), {"__call__": provided, "__bool__": lambda s: False})()
    )

    return prop


def list(**kwargs):
    """
    Create a configuration attribute that holds a list of configuration values.

    Best used only for configuration nodes. Use an :func:`attr` in combination with a
    :func:`types.list <.config.types.list>` type for simple values.
    """
    return ConfigurationListAttribute(**kwargs)


def dict(**kwargs):
    """
    Create a configuration attribute that holds a key value pairs of configuration values.

    Best used only for configuration nodes. Use an :func:`attr` in combination
    with a :func:`types.dict <.config.types.dict>` type for simple values.
    """
    return ConfigurationDictAttribute(**kwargs)


def catch_all(**kwargs):
    """
    Catches any unknown key with a value that can be cast to the given type and collects
    them under the attribute name.
    """
    return ConfigurationAttributeCatcher(**kwargs)


def unset():
    """
    Override and unset an inherited configuration attribute.
    """
    return ConfigurationAttribute(unset=True)


def file(**kwargs):
    """
    Create a file dependency attribute.
    """
    from ..storage._files import FileDependencyNode

    kwargs.setdefault("type", FileDependencyNode)
    return attr(**kwargs)


def _setattr(instance, name, value):
    instance.__dict__["_" + name] = value


def _getattr(instance, name):
    try:
        return instance.__dict__["_" + name]
    except KeyError as e:
        instance.__getattribute__(e.args[0])


def _hasattr(instance, name):
    return "_" + name in instance.__dict__


def _get_root(obj):
    parent = obj
    while hasattr(parent, "_config_parent") and parent._config_parent is not None:
        parent = parent._config_parent
    return parent


def _strict_root(obj):
    root = _get_root(obj)
    return root if getattr(root, "_config_isroot", False) else None


def _booted_root(obj):
    root = _strict_root(obj)
    return root if root and root._config_isfinished and root._config_isbooted else None


def _is_booted(obj):
    return obj and obj._config_isbooted


def _root_is_booted(obj):
    root = _strict_root(obj)
    return _is_booted(root)


def _boot_nodes(top_node, scaffold):
    for node in walk_nodes(top_node):
        node.scaffold = scaffold
        # Boot attributes
        for attr in getattr(node, "_config_attrs", {}).values():
            booted = {None}
            for cls in type(node).__mro__:
                cls_attr = getattr(cls, attr.attr_name, None)
                if (boot := getattr(cls_attr, "__boot__", None)) and boot not in booted:
                    boot(node, scaffold)
                    booted.add(boot)
        # Boot node hook
        try:
            run_hook(node, "boot")
        except Exception as e:
            errr.wrap(BootError, e, prepend=f"Failed to boot {node}:")
    # fixme: why is this here? Will deadlock in case of BootError on specific node only.
    scaffold._comm.barrier()


def _unset_nodes(top_node):
    for node in walk_nodes(top_node):
        with contextlib.suppress(Exception):
            del node.scaffold
        node._config_parent = None
        node._config_key = None
        if hasattr(node, "_config_index"):
            node._config_index = None
        run_hook(node, "unboot")


class ConfigurationAttribute:
    """
    Base implementation of all the different configuration attributes.

    Call the factory
    function :func:`.attr` instead.
    """

    def __init__(
        self,
        type=None,
        default=None,
        call_default=None,
        required=False,
        key=False,
        unset=False,
        hint=MISSING,
    ):
        if not callable(required):
            self.required = lambda s: required
        else:
            self.required = required
        self.key = key
        self.default = default
        self.call_default = call_default
        self.type = self._set_type(type, key)
        self.unset = unset
        self.hint = hint

    def __set_name__(self, owner, name):
        self.attr_name = name

    def __get__(self, instance, owner):
        if instance is None:
            return self
        return _getattr(instance, self.attr_name)

    def fset(self, instance, value):
        return _setattr(instance, self.attr_name, value)

    def __set__(self, instance, value):
        if _hasattr(instance, self.attr_name):
            ex_value = _getattr(instance, self.attr_name)
            _unset_nodes(ex_value)
        if value is None:
            # Don't try to cast None to a value of the attribute type.
            return self.fset(instance, None)
        try:
            value = self.type(value, _parent=instance, _key=self.attr_name)
        except ValueError:
            # This value error should only arise when users are manually setting
            # attributes in an already bootstrapped config tree.
            raise CastError(
                f"'{value}' is not convertible to {self.type.__name__},"
                f" for attribute '{self.attr_name}' of {instance}."
            ) from None
        except (RequirementError, CastError) as e:
            if not hasattr(e, "node") or not e.node:
                e.node, e.attr = instance, self.attr_name
            raise
        except Exception as e:
            raise CastError(
                f"Couldn't cast '{value}' into {self.type.__name__}: {e}",
                instance,
                self.attr_name,
            ) from e
        self.flag_dirty(instance)
        # The value was cast to its intended type and the new value can be set.
        self.fset(instance, value)
        root = _strict_root(instance)
        if _is_booted(root):
            _boot_nodes(value, root.scaffold)

    def _set_type(self, type, key):
        self._config_type = type
        # Determine type of the attribute
        if not type and self.default is not None:
            if self.should_call_default():
                t = builtins.type(self.default())
            else:
                t = builtins.type(self.default)
        else:
            from . import types

            t = type or (key and types.key()) or types.str()
        # This call wraps the type handler so that it accepts all reserved keyword args
        # like `_parent` and `_key`
        t = _wrap_reserved(t)
        return t

    def get_type(self):
        return self._config_type

    def get_hint(self):
        if self.hint is not MISSING:
            return self.hint
        if hasattr(self.type, "__hint__"):
            return self.type.__hint__()
        return MISSING

    def get_node_name(self, instance):
        return instance.get_node_name() + "." + self.attr_name

    def is_node_type(self):
        return hasattr(self._config_type, "_config_attrs")

    def tree(self, instance):
        val = _getattr(instance, self.attr_name)
        return self.tree_of(val)

    def tree_of(self, value):
        # Allow subnodes and other class values to convert themselves to their tree
        # representation
        if hasattr(value, "__tree__"):
            value = value.__tree__()
        # Check if the type handler specifies any inversion function to convert tree
        # values back to how they were found in the document.
        if hasattr(self.type, "__inv__") and value is not None:
            value = self.type.__inv__(value)
        return value

    def flag_dirty(self, instance):
        instance._config_state[self.attr_name] = False
        if self.attr_name not in instance._config_attr_order:
            instance._config_attr_order.append(self.attr_name)

    def is_dirty(self, instance):
        return not instance._config_state.get(self.attr_name, True)

    def flag_pristine(self, instance):
        instance._config_state[self.attr_name] = True

    def get_default(self):
        return self.default() if self.should_call_default() else self.default

    def should_call_default(self):
        cdf = self.call_default
        return cdf or (cdf is None and callable(self.default))


class cfglist(builtins.list):
    """
    Extension of the builtin list to manipulate lists of configuration nodes.
    """

    def get_node_name(self):
        return self._config_parent.get_node_name() + "." + self._config_attr_name

    def append(self, item):
        item = self._preset(len(self), item)
        super().append(item)
        self._postset((item,))
        return self[-1]

    def insert(self, index, item):
        item = self._preset(index, item)
        super().insert(index, item)
        self._postset((item,))
        self._reindex(index)

    def pop(self, index=-1):
        ex_item = super().pop(index)
        _unset_nodes(ex_item)
        self._reindex(index)
        return ex_item

    def clear(self):
        for node in self:
            _unset_nodes(node)
        super().clear()

    def sort(self, **kwargs):
        super().sort(**kwargs)
        self._reindex(0)

    def reverse(self):
        super().reverse()
        self._reindex(0)

    def extend(self, items):
        items = self._fromiter(len(self), items)
        super().extend(items)
        self._postset(items)

    def __setitem__(self, index, item):
        if isinstance(index, int):
            ex_items = [self[index]]
            item = self._preset(index, item)
            items = [item]
            reindex_from = None
        else:
            ex_items = self[index]
            reindex_from = index.indices(len(self))[0]
            items = item = self._fromiter(reindex_from, item)
        for ex_item in ex_items:
            _unset_nodes(ex_item)
        # Don't be fooled, item can be a single value or a list, depending on the index.
        super().__setitem__(index, item)
        if reindex_from is not None:
            self._reindex(reindex_from)
        self._postset(items)

    def _reindex(self, start):
        for i in range(start, len(self)):
            self[i]._config_key = i
            self[i]._config_index = i

    def _fromiter(self, start, items):
        return tuple(self._preset(start + i, item) for i, item in enumerate(items))

    def _preset(self, index, item):
        try:
            item = self._elem_type(item, _parent=self, _key=index)
            with contextlib.suppress(Exception):
                item._config_index = index
            return item
        except (RequirementError, CastError) as e:
            e.args = (
                f"Couldn't cast element {index} from '{item}'"
                + f" into a {self._elem_type.__name__}. "
                + e.msg,
                *e.args,
            )
            if not e.node:
                e.node, e.attr = self, index
            raise
        except Exception as e:
            raise CastError(
                f"Couldn't cast element {index} from '{item}'"
                + f" into a {self._elem_type.__name__}: {e}"
            ) from None

    def _postset(self, items):
        root = _strict_root(self)
        if root is not None:
            for item in items:
                _resolve_references(root, item)
        if _is_booted(root):
            for item in items:
                _boot_nodes(item, root.scaffold)

    @builtins.property
    def _config_attr_name(self):
        return self._config_attr.attr_name


class ConfigurationListAttribute(ConfigurationAttribute):
    def __init__(self, *args, size=None, **kwargs):
        super().__init__(*args, **kwargs)
        self.size = size

    def __set__(self, instance, value, _key=None):
        _setattr(instance, self.attr_name, self.fill(value, _parent=instance))
        self.flag_dirty(instance)

    def __populate__(self, instance, value, unique_list=False):
        cfglist = _getattr(instance, self.attr_name)
        if not unique_list or value not in cfglist:
            builtins.list.append(cfglist, value)

    def fill(self, value, _parent, _key=None):
        _cfglist = cfglist()
        _cfglist._config_parent = _parent
        _cfglist._config_attr = self
        _cfglist._elem_type = self.child_type
        if isinstance(value, builtins.dict):
            raise CastError(f"Dictionary `{value}` given where list is expected.")
        _cfglist.extend(value or builtins.list())
        if self.size is not None and len(_cfglist) != self.size:
            raise CastError(
                f"Couldn't cast {value} into a {self.size}-element list,"
                + f" obtained {len(_cfglist)} elements"
            )
        return _cfglist

    def _set_type(self, type, key=None):
        self.child_type = super()._set_type(type, key=False)

        @wraps(self.fill)
        def wrapper(*args, **kwargs):
            return self.fill(*args, **kwargs)

        # Expose children __inv__ function if it exists
        if hasattr(self.child_type, "__inv__"):
            wrapper.__inv__ = self.child_type.__inv__
        return wrapper

    def tree(self, instance):
        val = _getattr(instance, self.attr_name)
        return [self.tree_of(e) for e in val]

    def get_hint(self):
        if self.hint is not MISSING:
            return self.hint
        if hasattr(self.child_type, "__hint__"):
            return [self.child_type.__hint__(), self.child_type.__hint__()]
        return MISSING


class cfgdict(builtins.dict):
    """
    Extension of the builtin dictionary to manipulate dicts of configuration nodes.
    """

    def __getattr__(self, name):
        try:
            return self[name]
        except KeyError:
            raise AttributeError(
                self.get_node_name() + f" object has no attribute '{name}'"
            ) from None

    def __setitem__(self, key, value):
        if key in self:
            _unset_nodes(self[key])
        try:
            value = self._elem_type(value, _parent=self, _key=key)
        except (RequirementError, CastError) as e:
            if not (hasattr(e, "node") and e.node):
                e.node, e.attr = self, key
            raise
        except Exception:
            import traceback

            raise CastError(
                f"Couldn't cast {self.get_node_name()}.{key} from '{value}' "
                + f"into a {self._elem_type.__name__}\n"
                + traceback.format_exc()
            ) from None
        else:
            super().__setitem__(key, value)
            root = _strict_root(value)
            if root is not None:
                _resolve_references(root, value)
            if _is_booted(root):
                _boot_nodes(value, root.scaffold)

    def add(self, key, *args, **kwargs):
        if key in self:
            raise KeyError(
                f"{self.get_node_name()} already contains '{key}'."
                + " Use `node[key] = value` if you want to overwrite it."
            )
        self._config_attr.flag_dirty(self._config_parent)
        self[key] = value = self._elem_type(*args, _parent=self, _key=key, **kwargs)
        return value

    def clear(self):
        for node in self.values():
            _unset_nodes(node)
        super().clear()

    def pop(self, key):
        item = super().pop(key)
        _unset_nodes(item)
        return item

    def popitem(self):
        key, value = super().popitem()
        _unset_nodes(value)
        return key, value

    def setdefault(self, key, value):
        if key in self:
            return self[key]
        else:
            self[key] = value
            return self[key]

    def update(self, other):
        for ckey, value in other.items():
            self[ckey] = value

    def __ior__(self, other):
        ex_values = tuple(self.values())
        try:
            merge_f = super().__ior__
        except AttributeError:
            # Patch for 3.8
            merge_f = super().update
        merge_f(other)
        new_values = tuple(self.values())
        for removed_node in (e for e in ex_values if e not in new_values):
            _unset_nodes(removed_node)
        for added_node in (a for a in new_values if a not in ex_values):
            _boot_nodes(added_node, self.scaffold)
        return self

    def copy(self):
        return cfgdictcopy(self)

    @builtins.property
    def _config_attr_name(self):
        return self._config_attr.attr_name

    def get_node_name(self):
        return self._config_parent.get_node_name() + "." + self._config_attr_name


class cfgdictcopy(builtins.dict):
    def __init__(self, other):
        super().__init__(other)
        self._elem_type = other._elem_type
        self._copied_from = other

    @builtins.property
    def _config_attr_name(self):
        return self._copied_from._config_attr_name

    def get_node_name(self):
        return self._copied_from.get_node_name()


class ConfigurationDictAttribute(ConfigurationAttribute):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def __set__(self, instance, value, _key=None):
        _setattr(
            instance,
            self.attr_name,
            self.fill(value, _parent=instance, _key=_key or self.attr_name),
        )

    def fill(self, value, _parent, _key=None):
        _cfgdict = cfgdict()
        _cfgdict._config_parent = _parent
        _cfgdict._config_key = _key
        _cfgdict._config_attr = self
        _cfgdict._elem_type = self.child_type
        _cfgdict.update(value or builtins.dict())
        return _cfgdict

    def _set_type(self, type, key=None):
        self.child_type = super()._set_type(type, key=False)

        @wraps(self.fill)
        def wrapper(*args, **kwargs):
            return self.fill(*args, **kwargs)

        # Expose children __inv__ function if it exists
        if hasattr(self.child_type, "__inv__"):
            wrapper.__inv__ = self.child_type.__inv__
        return wrapper

    def tree(self, instance):
        val = _getattr(instance, self.attr_name).items()
        return {k: self.tree_of(v) for k, v in val}

    def get_hint(self):
        if self.hint is not MISSING:
            return self.hint
        if hasattr(self.child_type, "__hint__"):
            return {
                "key1": self.child_type.__hint__(),
                "key2": self.child_type.__hint__(),
            }
        return MISSING


class ConfigurationReferenceAttribute(ConfigurationAttribute):
    def __init__(
        self,
        reference,
        key=None,
        ref_type=None,
        populate=None,
        backref=None,
        pop_unique=True,
        **kwargs,
    ):
        self.ref_lambda = reference
        self.ref_key = key
        self.ref_type = ref_type
        self.populate = populate
        self.backref = backref
        self.pop_unique = pop_unique
        # No need to cast to any types: the reference we fetch will already have been cast
        if "type" in kwargs:  # pragma: nocover
            del kwargs["type"]
        super().__init__(**kwargs)

    def get_ref_key(self):
        return self.ref_key or (self.attr_name + "_reference")

    def __set__(self, instance, value, key=None):
        if self.is_reference_value(value):
            _setattr(instance, self.attr_name, value)
        else:
            setattr(instance, self.get_ref_key(), value)
            if self.should_resolve_on_set(instance):
                if hasattr(instance, "_config_root"):  # pragma: nocover
                    raise CfgReferenceError(
                        "Can't autoresolve references without a config root."
                    )
                _setattr(
                    instance,
                    self.attr_name,
                    self.__ref__(instance, instance._config_root),
                )

    def is_reference_value(self, value):
        if value is None:
            return True
        if self.ref_type is not None:
            return isinstance(value, self.ref_type)
        elif hasattr(self.ref_lambda, "is_ref"):
            return self.ref_lambda.is_ref(value)
        else:
            return not isinstance(value, str)

    def should_resolve_on_set(self, instance):
        return (
            hasattr(instance, "_config_resolve_on_set")
            and instance._config_resolve_on_set
        )

    def __ref__(self, instance, root):
        try:
            remote, remote_key = self._prepare_self(instance, root)
        except NoReferenceAttributeSignal:
            return None
        return self.resolve_reference(instance, remote, remote_key)

    def _prepare_self(self, instance, root):
        instance._config_root = root
        instance._config_resolve_on_set = True
        remote = self.ref_lambda(root, instance)
        local_attr = self.get_ref_key()
        if not hasattr(instance, local_attr):
            raise NoReferenceAttributeSignal()
        return remote, getattr(instance, local_attr)

    def resolve_reference(self, instance, remote, key):
        if key not in remote:
            raise CfgReferenceError(
                f"Reference '{key}' of {self.get_node_name(instance)} "
                f"does not exist in {remote.get_node_name()}"
            )
        value = remote[key]
        if self.populate:
            self.populate_reference(instance, value)
        if self.backref:
            self.back_reference(instance, value)

        return value

    def populate_reference(self, instance, reference):
        # Remote descriptors can ask to handle populating itself by implementing a
        # __populate__ method. Here we check if the method exists and if so defer to it.
        if (pop_attr := getattr(reference.__class__, self.populate, None)) and (
            pop_func := getattr(pop_attr, "__populate__", None)
        ):
            pop_func(reference, instance, unique_list=self.pop_unique)
        elif (population := getattr(reference, self.populate, None)) is not None:
            if not self.pop_unique or instance not in population:
                population.append(instance)
        else:
            setattr(reference, self.populate, [instance])

    def back_reference(self, instance, reference):
        if (ref_attr := getattr(reference.__class__, self.backref, None)) and (
            ref_func := getattr(ref_attr, "__backref__", None)
        ):
            ref_func(reference, instance)
        else:
            setattr(reference, self.backref, instance)

    def tree(self, instance):
        val = getattr(instance, self.get_ref_key(), None)
        if self.is_reference_value(val) and hasattr(val, "_config_key"):
            val = val._config_key
        return val


class ConfigurationReferenceListAttribute(ConfigurationReferenceAttribute):
    def __set__(self, instance, value, key=None):
        if value is None:
            setattr(instance, self.get_ref_key(), [])
            _setattr(instance, self.attr_name, [])
            return
        try:
            remote_keys = builtins.list(iter(value))
        except TypeError:
            raise CfgReferenceError(
                f"Reference list '{value}' of "
                f"{self.get_node_name(instance)} is not iterable."
            ) from None
        # Store the referring values to the references key.
        setattr(instance, self.get_ref_key(), remote_keys)
        if self.should_resolve_on_set(instance):
            remote = self.ref_lambda(instance._config_root, instance)
            refs = self.resolve_reference_list(instance, remote, remote_keys)
            _setattr(instance, self.attr_name, refs)

    def __get__(self, instance, owner):
        if instance is None:
            return self
        if self.should_resolve_on_set(instance):
            return super().__get__(instance, owner)
        else:
            return getattr(instance, self.get_ref_key())

    def get_ref_key(self):
        return self.ref_key or (self.attr_name + "_references")

    def __ref__(self, instance, root):
        try:
            remote, remote_keys = self._prepare_self(instance, root)
        except NoReferenceAttributeSignal:  # pragma: nocover
            return None
        return self.resolve_reference_list(instance, remote, remote_keys)

    def resolve_reference_list(self, instance, remote, remote_keys):
        refs = []
        for remote_key in remote_keys:
            if not self.is_reference_value(remote_key):
                reference = self.resolve_reference(instance, remote, remote_key)
            else:
                reference = remote_key
                # Usually resolve_reference also populates, but since we have our ref
                # already we skip it and should call populate_reference ourselves.
                if self.populate:
                    self.populate_reference(instance, reference)
            refs.append(reference)
        return refs

    def __populate__(self, instance, value, unique_list=False):
        has_refs = hasattr(instance, self.get_ref_key())
        has_pop = hasattr(instance, self.attr_name)
        if has_pop:
            population = getattr(instance, self.attr_name)
        if has_refs:
            references = getattr(instance, self.get_ref_key())
        is_new = (not has_pop or value not in population) and (
            not has_refs or value not in references
        )
        should_pop = has_pop and (not unique_list or is_new)
        should_ref = should_pop or not has_pop
        if should_pop:
            population.append(value)
        if should_ref:
            if not has_refs:
                setattr(instance, self.get_ref_key(), [value])
            else:
                references.append(value)

    def tree(self, instance):
        val = getattr(instance, self.get_ref_key(), [])
        val = [v._config_key if self._tree_should_unreference(v) else v for v in val]
        return val

    def _tree_should_unreference(self, value):
        return self.is_reference_value(value) and hasattr(value, "_config_key")


class ConfigurationAttributeSlot(ConfigurationAttribute):
    def __set__(self, instance, value):  # pragma: nocover
        raise NotImplementedError(
            f"Configuration slot '{self.attr_name}' of {instance.get_node_name()} is"
            f" empty. The {instance.__class__._bsb_entry_point.module_name} plugin"
            f" provided by '{instance.__class__._bsb_entry_point.dist}' should fill the"
            " slot with a configuration attribute."
        )


class ConfigurationProperty(ConfigurationAttribute):
    def __init__(self, fget, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.fget = fget
        self.fset = None

    def setter(self, f):
        self.fset = f
        return self

    def __get__(self, instance, owner):
        if instance is None:
            return self
        return self.fget(instance)

    def __set__(self, instance, value):
        if self.fset is None:
            e = AttributeError(f"Can't set attribute '{self.attr_name}'")
            e.node = self
            raise e
        else:
            return super().__set__(instance, value)


def _collect_kv(n, d, k, v):
    d[k] = v


class ConfigurationAttributeCatcher(ConfigurationAttribute):
    def __init__(
        self,
        *args,
        type=str,
        initial=builtins.dict,
        catch=_collect_kv,
        contains=None,
        tree_cb=None,
        **kwargs,
    ):
        super().__init__(*args, type=type, default=initial, call_default=True, **kwargs)
        self.catch_callback = catch
        if contains is not None:
            self.contains = contains
        if tree_cb is not None:
            self.tree_callback = tree_cb

    def __set__(self, instance, value):
        _setattr(instance, self.attr_name, value)

    def get_caught(self, instance):
        if not hasattr(instance, f"_{self.attr_name}_caught"):
            setattr(instance, f"_{self.attr_name}_caught", {})
        return getattr(instance, f"_{self.attr_name}_caught")

    def __catch__(self, node, key, value):
        # Try to cast to our type, if it fails it will be caught by whoever is asking us
        # to catch this and we don't catch this value.
        cast = self.type(value, _parent=node, _key=key)
        # If succesfully cast, catch this value by executing our catch callback.
        self.catch_callback(node, _getattr(node, self.attr_name), key, cast)
        self.get_caught(node)[key] = cast

    def tree(self, instance):
        # The default attr catcher collects what it catches in a dict. When we want to
        # build the config tree again these values should be placed back in their
        # original keys. We don't want to store our caught values in the config file. To
        # do so we use the `tree_callback` instead.
        return None

    def contains(self, instance, key):
        return key in self.get_caught(instance)

    def tree_callback(self, instance, key):
        # When building the config tree the values that were caught can't be found in the
        # attrs and the tree builder will check all catch-attr's `contains` methods and
        # calls the right tree_callback to fetch the value.
        value = _getattr(instance, self.attr_name)[key]
        if hasattr(value, "__tree__"):
            value = value.__tree__()
        return value



"""
Service module.

Register or access interfaces that may be provided, mocked or missing, but should always
behave neatly on import.
"""

from ._util import ErrorModule as _ErrorModule
from .mpi import MPIService as _MPIService
from .mpilock import MPILockModule as _MPILockModule

MPI = _MPIService()
"""
MPI service.
"""
MPILock = _MPILockModule("mpilock")
"""
MPILock service.
"""

from .pool import JobPool as _JobPool  # noqa E402  # needs to be imported after MPIService
from .pool import WorkflowError, pool_cache  # noqa E402  # needs to be imported after MPIService

JobPool = _JobPool
"""
JobPool service.
"""


def __getattr__(attr):
    return _ErrorModule(f"{attr} is not a registered service.")


def register_service(attr, provider):
    globals()[attr] = provider


__all__ = [
    "MPI",
    "MPILock",
    "JobPool",
    "register_service",
    "WorkflowError",
    "pool_cache",
]



from __future__ import annotations

import abc
import itertools
import typing
from collections import deque
from functools import reduce

import morphio
import numpy as np

from ... import config
from ..._encoding import EncodedLabels
from ...config import types
from .. import Morphology

if typing.TYPE_CHECKING:
    from ...storage._files import FileDependency


@config.dynamic(attr_name="parser", auto_classmap=True, default="bsb")
class MorphologyParser:
    cls: type = config.attr(type=types.class_(), default="bsb.morphologies.Morphology")
    branch_cls: type = config.attr(type=types.class_(), default="bsb.morphologies.Branch")

    @abc.abstractmethod
    def parse(self, file: FileDependency | str) -> Morphology:
        """
        Parse the morphology.
        """
        pass


@config.node
class BsbParser(MorphologyParser, classmap_entry="bsb"):
    tags: dict[str | list[str]] = config.attr(
        type=types.dict(type=types.or_(types.str(), types.list(str)))
    )
    """
    Dictionary mapping SWC tags to sets of morphology labels.
    """
    skip_boundary_labels: list[str] = config.attr(type=types.list(str))
    """
    A set of labels that is used to create gaps in a morphology at certain boundaries.

    No point will be inferred between a child branch of a branch labelled with the given
    labels; usually used to skip points between the soma and its child branches.
    """

    def parse(self, file: FileDependency | str):
        from ...storage._files import FileDependency

        if not isinstance(file, FileDependency):
            file = FileDependency(file)
        content, encoding = file.get_content(check_store=False)
        return self.parse_content(content.decode(encoding or "utf8"))

    def parse_content(self, content: str):
        data = self._swc_parse(content)
        return self._swc_data_to_morpho(data)

    def _swc_parse(self, content: str):
        try:
            data = [
                swc_data
                for line in content.split("\n")
                if not line.strip().startswith("#")
                and (swc_data := [float(x) for x in line.split() if x != ""])
            ]
        except Exception:
            raise RuntimeError("Could not parse SWC content") from None
        err_lines = ", ".join(str(i) for i, d in enumerate(data) if len(d) != 7)
        if err_lines:
            raise ValueError(f"SWC incorrect on lines: {err_lines}")
        return np.array(data)

    def _swc_data_to_morpho(self, data):
        data = np.array(data, copy=False)
        tag_map = {1: "soma", 2: "axon", 3: "dendrites"}
        if self.tags is not None:
            tag_map.update((int(k), v) for (k, v) in self.tags.items())
        # `data` is the raw SWC data,
        # `samples` and `parents` are the graph nodes and edges.
        samples = data[:, 0].astype(int)
        # Map possibly irregular sample IDs (SWC spec allows this)
        # to an ordered 0 to N map.
        id_map = dict(zip(samples, itertools.count()))
        id_map[-1] = -1
        # Create an adjacency list of the graph described in the SWC data
        adjacency = {n: [] for n in range(len(samples))}
        adjacency[-1] = []
        map_ids = np.vectorize(id_map.get)
        parents = map_ids(data[:, 6])
        for s, p in enumerate(parents):
            adjacency[p].append(s)
        # Now turn the adjacency list into a list of unbranching stretches of the graph.
        # Call these `node_branches` because they only contain the sample/node ids.
        node_branches = []
        for root_node in adjacency[-1]:
            self._swc_branch_dfs(adjacency, node_branches, root_node, data, tag_map)
        branches = []
        roots = []
        _len = sum(len(s[1]) for s in node_branches)
        points = np.empty((_len, 3))
        radii = np.empty(_len)
        tags = np.empty(_len, dtype=int)
        labels = EncodedLabels.none(_len)
        # Now turn each "node branch" into an actual branch by looking up the node data in
        # the samples array. We copy over the node data into several contiguous matrices
        # that will form the basis of the Morphology data structure.
        ptr = 0
        for parent, branch_nodes in node_branches:
            node_data = data[branch_nodes]
            nptr = ptr + len(node_data)
            # Example with the points data matrix:
            # copy over the swc data into contiguous arr
            points[ptr:nptr] = node_data[:, 2:5]
            # Then create a partial view into that data matrix for the branch
            branch_points = points[ptr:nptr]
            # Same here for radius,
            radii[ptr:nptr] = node_data[:, 5]
            branch_radii = radii[ptr:nptr]
            # the SWC tags
            tags[ptr:nptr] = node_data[:, 1]
            if len(branch_nodes) > 1:
                # Since we add an extra point we have to copy its tag from the next point.
                tags[ptr] = tags[ptr + 1]
            branch_tags = tags[ptr:nptr]
            # And the labels
            branch_labels = labels[ptr:nptr]
            for v in np.unique(branch_tags):
                u_tags = tag_map.get(v, f"tag_{v}")
                branch_labels.label(
                    [u_tags] if isinstance(u_tags, str) else u_tags, branch_tags == v
                )
            ptr = nptr
            # Use the views to construct the branch
            branch = self.branch_cls(branch_points, branch_radii, branch_labels)
            branch.set_properties(tags=branch_tags)
            branches.append(branch)
            if parent is not None:
                branches[parent].attach_child(branch)
            else:
                roots.append(branch)
        # Then save the shared data matrices on the morphology
        morpho = self.cls(roots, shared_buffers=(points, radii, labels, {"tags": tags}))
        # And assert that this shared buffer mode succeeded
        assert morpho._check_shared(), "SWC import didn't result in shareable buffers."
        return morpho

    def _swc_branch_dfs(self, adjacency, branches, node, data, tags):
        boundaries = set()
        if self.skip_boundary_labels:
            tset = set(self.skip_boundary_labels)
            for tag, labels in tags.items():
                lset = set(labels if not isinstance(labels, str) else [labels])
                if tset.issuperset(lset):
                    boundaries.add(tag)
        branch = []
        branch_id = len(branches)
        branches.append((None, branch))
        node_stack = deque()
        while True:
            if node is not None:
                # Append the current node to the current branch, and get its child nodes
                branch.append(node)
                child_nodes = adjacency[node]

            if not child_nodes:
                # No children, pop next branch
                try:
                    parent_bid, parent, node, skip = node_stack.pop()
                except IndexError:
                    # No next branch, we're done
                    break
                else:
                    # Start the next branch
                    branch = [] if skip else [parent]
                    branch_id = len(branches)
                    branches.append((parent_bid, branch))
            elif len(child_nodes) == 1 and not (
                data[node, 1] in boundaries and data[child_nodes[0], 1] not in boundaries
            ):
                # One child, and not a skipped boundary: grow the branch
                node = child_nodes[0]
            else:
                # Branch point: create 1 new branch per child point
                # If skip is False we add the current node to all the child branches.
                skip = data[node, 1] in boundaries
                node_stack.extend(
                    (branch_id, node, child, skip) for child in reversed(child_nodes)
                )
                child_nodes = []
                node = None


# Wrapper to append our own attributes to morphio somas and treat it like any other branch
class _MorphIoSomaWrapper:
    def __init__(self, obj):
        self._o = obj

    def __getattr__(self, attr):
        return getattr(self._o, attr)


@config.node
class MorphIOParser(MorphologyParser, classmap_entry="morphio"):
    @config.property(type=types.list(type=types.in_(morphio.Option.__members__)))
    def flags(self):
        return getattr(self, "_flags", morphio.Option.no_modifier)

    @flags.setter
    def flags(self, values):
        self._flags = reduce(
            morphio.Option.__or__,
            [getattr(morphio.Option, flag) for flag in values or []],
            morphio.Option.no_modifier,
        )

    def parse(self, file: FileDependency | str) -> Morphology:
        from ...storage._files import FileDependency

        if isinstance(file, str):
            file = FileDependency(file)

        with file.provide_locally() as (fp, encoding):
            morpho_io = morphio.Morphology(fp, self.flags)
        # We create shared buffers for the entire morphology, which optimize operations on
        # the entire morphology such as `.flatten`, subtree transformations and IO.
        # The branches have views on those buffers, and as long as no points are added or
        # removed, we can keep working in shared buffer mode.
        soma = _MorphIoSomaWrapper(morpho_io.soma)
        _len = len(morpho_io.points) + len(soma.points)
        points = np.empty((_len, 3))
        radii = np.empty(_len)
        tags = np.empty(_len, dtype=int)
        labels = EncodedLabels.none(_len)
        soma.children = morpho_io.root_sections
        section_stack = deque([(None, soma)])
        branch = None
        roots = []
        ptr = 0
        while True:
            try:
                parent, section = section_stack.pop()
            except IndexError:
                break
            else:
                nptr = ptr + len(section.points)
                # Fill the branch data into the shared buffers and create views into them.
                points[ptr:nptr] = section.points
                branch_points = points[ptr:nptr]
                radii[ptr:nptr] = section.diameters / 2
                branch_radii = radii[ptr:nptr]
                tags[ptr:nptr] = np.ones(len(section.points), dtype=int) * int(
                    section.type
                )
                branch_tags = tags[ptr:nptr]
                branch_labels = labels[ptr:nptr]
                ptr = nptr
                # Pass the shared buffer views to the branch
                branch = self.branch_cls(branch_points, branch_radii, branch_labels)
                branch.set_properties(tags=branch_tags)
                if parent:
                    parent.attach_child(branch)
                else:
                    roots.append(branch)
                children = reversed([(branch, child) for child in section.children])
                section_stack.extend(children)
        morpho = self.cls(roots, shared_buffers=(points, radii, labels, {"tags": tags}))
        assert morpho._check_shared(), (
            "MorphIO import didn't result in shareable buffers."
        )
        return morpho


__all__ = ["BsbParser", "MorphIOParser", "MorphologyParser"]



from exceptiongroup import ExceptionGroup

from .exceptions import PackageRequirementWarning
from .reporting import warn


class MissingRequirementErrors(ExceptionGroup):
    pass


def get_missing_requirement_reason(package):
    from importlib.metadata import PackageNotFoundError, version

    from packaging.requirements import InvalidRequirement, Requirement

    try:
        req = Requirement(package)
    except InvalidRequirement:
        return f"Can't check package requirement '{package}': invalid requirement"
    try:
        ver = version(req.name)
    except PackageNotFoundError:
        return (
            f"Missing package '{req.name}'. "
            "You may experience errors or differences in results."
        )
    else:
        if ver not in req.specifier:
            return (
                f"Installed version of '{req.name}' ({ver}) "
                f"does not match requirements: '{req}'. "
                "You may experience errors or differences in results."
            )


def get_missing_packages(packages):
    return [
        package
        for package in packages
        if get_missing_requirement_reason(package) is not None
    ]


def get_unmet_package_reasons(packages):
    return [
        reason
        for package in packages
        if (reason := get_missing_requirement_reason(package)) is not None
    ]


def warn_missing_packages(packages):
    for warning in get_unmet_package_reasons(packages):
        warn(warning, PackageRequirementWarning)


def raise_missing_packages(packages):
    raise MissingRequirementErrors(
        "Your model is missing requirement(s)",
        [
            PackageRequirementWarning(warning)
            for warning in get_unmet_package_reasons(packages)
        ],
    )



import abc
import contextlib
import datetime
import time
from collections import defaultdict
from typing import cast

from blessed import Terminal
from dashing import Color, DoubleColumn, HSplit, Text

from . import MPI
from .pool import (
    Job,
    JobStatus,
    PoolJobAddedProgress,
    PoolJobUpdateProgress,
    PoolProgress,
    PoolProgressReason,
    PoolStatus,
    Workflow,
)


class Listener(abc.ABC):
    @abc.abstractmethod
    def __call__(self, progress: PoolProgress):
        pass


class NonTTYTerminalListener(Listener):
    def __call__(self, progress: PoolProgress):
        if progress.reason == PoolProgressReason.JOB_STATUS_CHANGE:
            job = cast(PoolJobUpdateProgress, progress).job
            print(f"[{datetime.datetime.now()} - BSB] {job.status} {job.name}")
        if progress.reason == PoolProgressReason.MAX_TIMEOUT_PING:
            print(f"[{datetime.datetime.now()}] Progress ping.")


class TTYTerminalListener(Listener):
    def __init__(self, fps: int):
        self._workflow: Workflow = None
        self._terminal = None
        self._last_update = 0
        self.fps = fps

        self._ui = HSplit(
            Text(
                "",
                title="Progress",
                color=Color.White,
                border_color=Color.Yellow,
            ),
            DoubleColumn(
                [],
                title="Jobs",
                color=7,
                border_color=3,
            ),
        )

    @contextlib.contextmanager
    def _terminal_context(self, terminal: Terminal):
        with terminal.fullscreen(), terminal.cbreak(), terminal.hidden_cursor():
            yield

    def __enter__(self):
        if not MPI.get_rank():
            self._terminal = Terminal()
            self._context = self._terminal_context(self._terminal)
            self._ui._terminal = self._terminal
            self._context.__enter__()
            self._start = time.perf_counter()
            self.update()

    def __exit__(self, exc_type, exc_val, exc_tb):
        if not MPI.get_rank():
            self._context.__exit__(exc_type, exc_val, exc_tb)

    def _initialize(self, progress: PoolProgress):
        self._tally = PoolTally(JobTally)
        if progress.workflow:
            self._workflow = progress.workflow

    def __call__(self, progress: PoolProgress):
        if MPI.get_rank():
            return
        if progress.reason == PoolProgressReason.POOL_STATUS_CHANGE:
            if progress.status == PoolStatus.SCHEDULING:
                self._initialize(progress)
            self._pool_status = progress.status
        if progress.reason == PoolProgressReason.JOB_ADDED:
            job = cast(PoolJobAddedProgress, progress).job
            self._tally[job.name].tally(job)
            self.update_jobs()
        self.update_progress()
        if (
            (progress.reason == PoolProgressReason.JOB_STATUS_CHANGE)
            and progress.status != JobStatus.PENDING
            and progress.status != JobStatus.QUEUED
        ):
            self._tally.update_tally(cast(PoolJobUpdateProgress, progress))
            self.update_jobs()
        self.update()

    def update(self):
        if time.time() - self._last_update >= 1 / self.fps:
            self._ui.display()
            self._last_update = time.time()

    def update_jobs(self):
        cast(DoubleColumn, self._ui.items[1]).lines = [
            (name, str(tally)) for name, tally in self._tally.top()
        ]

    def update_progress(self):
        cast(Text, self._ui.items[0]).text = (
            (f"Workflow: {' | '.join(self._workflow.phases)}\n" if self._workflow else "")
            + (f"Phase: {self._workflow.phase}\n" if self._workflow else "")
            + f"Status: {str(self._pool_status).split('.')[1]}\n"
            f"Job total:\n"
            f" {self._tally}\n"
            f"Elapsed: {round(time.perf_counter() - self._start, 2)}s\n"
            f"Job est.: {round(self._tally.get_estimate(), 2)}s"
        )


class JobTally:
    def __init__(self):
        self._total = 0
        self._tallies = defaultdict(int)

    @property
    def total(self):
        return self._total

    def tally(self, job: Job):
        self._tallies[job.status] += 1
        self._total += 1

    def tally_unfinished(self):
        return self.tally_status(JobStatus.PENDING, JobStatus.QUEUED, JobStatus.RUNNING)

    def tally_finished(self):
        return self._total - self.tally_unfinished()

    def tally_status(self, *status: JobStatus):
        return sum((self._tallies[s] for s in status), start=0)

    def progress(self):
        return self.tally_finished() / self._total

    def finished(self):
        return self.tally_unfinished() == 0

    def update(self, old_status: JobStatus, new_status: JobStatus):
        self._tallies[old_status] = max(0, self._tallies[old_status] - 1)
        self._tallies[new_status] += 1

    def __str__(self):
        others = ""
        if running := self.tally_status(JobStatus.RUNNING):
            others += f"→ {running}/"
        if failed := self.tally_status(JobStatus.FAILED, JobStatus.ABORTED):
            others += f"🗙 {failed}/"
        if skipped := self.tally_status(JobStatus.CANCELLED):
            others += f"⊘ {skipped}/"
        return f"({others}✔ {self.tally_status(JobStatus.SUCCESS)}/{self._total})"


class PoolTally(defaultdict):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._elapsed = {}
        self._durations = defaultdict(lambda: [0, 0])

    def sorted(self):
        yield from sorted(self.items(), key=lambda i: -i[1].progress())

    def top(self):
        yield from (i for i in self.sorted() if not i[1].finished())

    def update_tally(self, progress: PoolJobUpdateProgress):
        self[progress.job.name].update(progress.old_status, progress.status)
        if progress.status == JobStatus.RUNNING:
            self._elapsed[id(progress.job)] = time.perf_counter()
        elif progress.old_status == JobStatus.RUNNING:
            total, n = self._durations[progress.job.name]
            self._durations[progress.job.name] = [
                total + time.perf_counter() - self._elapsed[id(progress.job)],
                n + 1,
            ]
            del self._elapsed[id(progress.job)]

    def tally(self, *status: JobStatus):
        if status:
            return sum(tally.tally_status(*status) for tally in self.values())
        else:
            return sum(tally.total for tally in self.values())

    def get_estimate(self):
        return sum(
            total / n * self[job_name].tally_unfinished()
            for job_name, (total, n) in self._durations.items()
            if n
        )

    def __str__(self):
        others = ""
        if running := self.tally(JobStatus.RUNNING):
            others += f"→ {running}/"
        if failed := self.tally(JobStatus.FAILED, JobStatus.ABORTED):
            others += f"🗙 {failed}/"
        if skipped := self.tally(JobStatus.CANCELLED):
            others += f"⊘ {skipped}/"
        return f"({others}✔ {self.tally(JobStatus.SUCCESS)}/{self.tally()})"



from __future__ import annotations

from os import PathLike

from .parser import MorphologyParser


def parse_morphology_content(content: str | bytes, parser="bsb", **kwargs):
    return MorphologyParser(parser=parser, **kwargs).parse_content(content)


def parse_morphology_file(file: str | PathLike, parser="bsb", **kwargs):
    return MorphologyParser(parser=parser, **kwargs).parse(file)


__all__ = ["parse_morphology_content", "parse_morphology_file"]



import math

import numpy as np

from .. import config
from ..config import types
from ..exceptions import ConfigurationError, PackingError
from ..mixins import NotParallel
from ..reporting import report
from .strategy import PlacementStrategy


@config.node
class ParallelArrayPlacement(NotParallel, PlacementStrategy):
    """
    Implementation of the placement of cells in parallel arrays Cells are placed in rows
    on the plane defined by 2 selected axes.
    """

    spacing_x: float = config.attr(type=types.float(min=0), required=True)
    """
    Space in between two cells along the main axis.
    """
    angle: float = config.attr(type=types.deg_to_radian(), required=True)
    """
    Angle between the second axis and the axis of the rows of cells.
    """

    def boot(self):
        if self.angle % (np.pi) == np.pi / 2:
            raise ConfigurationError(
                "Parallel array angle should be not a multiple of pi/2 for "
                f"'{self.name}'. Provided angle: {self.angle}"
            )

    def place(self, chunk, indicators):
        """
        Cell placement: Create a lattice of parallel arrays/lines in the (x, y) surface.
        """
        for indicator in indicators.values():
            cell_type = indicator.cell_type
            radius = indicator.get_radius()
            for prt in self.partitions:
                width, depth, height = prt.data.mdc - prt.data.ldc
                ldc = prt.data.ldc
                # Add a random shift to the starting points of the arrays for variation.
                x_shift = np.random.rand() * self.spacing_x
                # Place cells equally spaced over the entire length of the X axis kept
                # apart by the provided space.
                # They are placed in straight lines, tilted by a certain angle by adding
                # a shifting value.
                x_pos = np.arange(start=0.0, stop=width, step=self.spacing_x) + x_shift
                if x_pos.shape[0] == 0:
                    # When the spacing_x of is larger than the simulation volume,
                    # place a single row on a random position along the X axis
                    x_pos = np.array([x_shift])
                # Amount of parallel arrays of cells
                n_arrays = x_pos.shape[0]
                # Number of cells
                n = np.sum(indicator.guess(prt.data))
                # Add extra cells to fill the lattice error volume which will be pruned
                n += int((n_arrays * self.spacing_x % width) / width * n)
                # cells to distribute along the rows
                cells_per_row = round(n / n_arrays)
                # The rounded amount of cells that will be placed
                cells_placed = cells_per_row * n_arrays
                # Calculate the position of the cells along the z-axis.
                y_pos, y_axis_distance = np.linspace(
                    start=0.0,
                    stop=depth - radius,
                    num=cells_per_row,
                    retstep=True,
                    endpoint=False,
                )
                # Center the cell soma center to the middle of the unit cell
                y_pos += radius + y_axis_distance / 2
                # The length of the X axis rounded up to a multiple of the unit cell size.
                lattice_x = n_arrays * self.spacing_x
                # The length of the X axis where cells can be placed in.
                bounded_x = lattice_x - radius * 2
                # Epsilon: open space in the unit cell along the Y axis
                epsilon = y_axis_distance / math.cos(self.angle) - radius * 2
                if epsilon < 0:
                    raise PackingError(
                        "Not enough space between cells placed on the same row "
                        f"for '{self.name}'."
                    )
                # Storage array for the cells
                cells = np.empty((cells_placed, 3))
                for i in range(y_pos.shape[0]):
                    # Shift the arrays at an angle
                    angleShift = y_pos[i] * math.tan(self.angle)
                    # Apply shift and offset
                    x = x_pos + angleShift
                    # Place the cells in a bounded lattice with a little modulus magic
                    x = ldc[0] + x % bounded_x + radius
                    # Place the cells in their y-position with jitter
                    y = (
                        ldc[1]
                        + y_pos[i]
                        + epsilon
                        * (np.random.rand(x.shape[0]) - 0.5)
                        * math.cos(self.angle)
                    )
                    # Place them at a uniformly random height throughout the partition.
                    z = ldc[2] + np.random.uniform(radius, height - radius, x.shape[0])
                    # Store this stack's cells
                    cells[(i * len(x)) : ((i + 1) * len(x)), 0] = x
                    cells[(i * len(x)) : ((i + 1) * len(x)), 1] = y
                    cells[(i * len(x)) : ((i + 1) * len(x)), 2] = z
                # Place all the cells in 1 batch (more efficient)
                positions = cells[cells[:, 0] < prt.data.ldc[0] + width - radius]

                # Determine in which chunks the cells must be placed
                cs = self.scaffold.configuration.network.chunk_size
                chunks_list = np.array(
                    [chunk.data + np.floor_divide(p, cs[0]) for p in positions]
                )
                unique_chunks_list = np.unique(chunks_list, axis=0)

                # For each chunk, place the cells
                for c in unique_chunks_list:
                    idx = np.where((chunks_list == c).all(axis=1))
                    pos_current_chunk = positions[idx]
                    self.place_cells(indicator, pos_current_chunk, chunk=c)
                report(f"Placed {len(positions)} {cell_type.name} in {prt.name}", level=3)


__all__ = ["ParallelArrayPlacement"]



from .. import config


@config.dynamic(attr_name="type", auto_classmap=True, required=False)
class ParameterValue:
    def __init__(self, value=None, /, **kwargs):
        self._constant = value


@config.dynamic(attr_name="type", auto_classmap=True, required=False)
class Parameter:
    value: ParameterValue = config.attr(type=ParameterValue)


__all__ = ["Parameter", "ParameterValue"]



from __future__ import annotations

import abc
import typing
from functools import cache
from itertools import chain

import numpy as np

from .. import config
from .._util import ichain, obj_str_insert
from ..config import refs, types
from ..exceptions import ConnectivityError
from ..mixins import HasDependencies
from ..profiling import node_meter
from ..reporting import warn
from ..storage._chunks import Chunk

if typing.TYPE_CHECKING:
    from ..cell_types import CellType
    from ..core import Scaffold
    from ..morphologies import MorphologySet
    from ..services import JobPool
    from ..storage.interfaces import PlacementSet


@config.node
class Hemitype:
    """
    Class used to represent one (pre- or postsynaptic) side of a connection rule.
    """

    scaffold: Scaffold

    cell_types: list[CellType] = config.reflist(refs.cell_type_ref, required=True)
    """
    List of cell types to use in connection.
    """
    labels: list[str] = config.attr(type=types.list())
    """
    List of labels to filter the placement set by.
    """
    morphology_labels: list[str] = config.attr(type=types.list())
    """
    List of labels to filter the morphologies by.
    """
    morpho_loader: typing.Callable[[PlacementSet], MorphologySet] = config.attr(
        type=types.function_(),
        required=False,
        call_default=False,
        default=(lambda ps: ps.load_morphologies()),
    )
    """
    Function to load the morphologies (MorphologySet) from a PlacementSet.

    This override can allow temporary dynamic morphology generation during the
    connectivity phase, from a much smaller, or empty, MorphologySet. It is useful for
    example when the task would take too much disk space or time otherwise.
    """

    def get_all_chunks(self):
        """
        Get the list of all chunks where the cell types were placed.

        :return: List of Chunks
        :rtype: list[bsb.storage._chunks.Chunk]
        """
        return [
            c for ct in self.cell_types for c in ct.get_placement_set().get_all_chunks()
        ]

    @cache
    def _get_rect_ext(self, chunk_size):
        # Returns the lower and upper boundary Chunk of the box containing the cell type
        # population, based on the cell type's morphology if it exists.
        # This box is centered on the Chunk [0., 0., 0.].
        # If no morphologies are associated to the cell types, the bounding box size is 0.
        types = self.cell_types
        loader = self.morpho_loader
        ps_list = [ct.get_placement_set() for ct in types]
        ms_list = [loader(ps) for ps in ps_list]
        if not sum(map(len, ms_list)):
            # No cells placed, return smallest possible RoI.
            return [np.array([0, 0, 0]), np.array([0, 0, 0])]
        metas = list(chain.from_iterable(ms.iter_meta(unique=True) for ms in ms_list))
        # TODO: Combine morphology extension information with PS rotation information.
        # Get the chunk coordinates of the boundaries of this chunk convoluted with the
        # extension of the intersecting morphologies.
        lbounds = np.min([m["ldc"] for m in metas], axis=0) // chunk_size
        ubounds = np.max([m["mdc"] for m in metas], axis=0) // chunk_size
        return lbounds, ubounds


class HemitypeCollection:
    """
    Class used to iterate over an ``Hemitype`` placement sets within a list of chunks, and
    over its cell types.
    """

    def __init__(self, hemitype: Hemitype, roi: list[Chunk]):
        self.hemitype = hemitype
        self.roi = roi

    def __iter__(self):
        return iter(self.hemitype.cell_types)

    @property
    def placement(self):
        """
        List the placement sets for each cell type, filtered according to the class
        morphology labels and list of chunks.

        :rtype: list[bsb.storage.interfaces.PlacementSet]
        """
        return [
            ct.get_placement_set(
                chunks=self.roi,
                labels=self.hemitype.labels,
                morphology_labels=self.hemitype.morphology_labels,
            )
            for ct in self.hemitype.cell_types
        ]


@config.dynamic(attr_name="strategy", required=True)
class ConnectionStrategy(abc.ABC, HasDependencies):
    scaffold: Scaffold
    name: str = config.attr(key=True)
    """
    Name used to refer to the connectivity strategy.
    """
    presynaptic: Hemitype = config.attr(type=Hemitype, required=True)
    """
    Presynaptic (source) neuron population.
    """
    postsynaptic: Hemitype = config.attr(type=Hemitype, required=True)
    """
    Postsynaptic (target) neuron population.
    """
    depends_on: list[ConnectionStrategy] = config.reflist(refs.connectivity_ref)
    """
    The list of strategies that must run before this one.
    """
    output_naming: str | None | dict[str, dict[str, str, None, list[str]]] = config.attr(
        type=types.or_(
            types.str(),
            types.dict(
                type=types.dict(
                    type=types.or_(
                        types.str(), types.list(type=types.str()), types.none()
                    )
                )
            ),
            types.list(type=types.str()),
        )
    )
    """
    Specifies how to name the output ConnectivitySets in which the connections between
    cell type pairs are stored.
    """

    def __init_subclass__(cls, **kwargs):
        super(cls, cls).__init_subclass__(**kwargs)
        # Decorate subclasses to measure performance
        node_meter("connect")(cls)

    def __hash__(self):
        return id(self)

    def __lt__(self, other):
        # This comparison should sort connection strategies by name, via __repr__ below
        return str(self) < str(other)

    @obj_str_insert
    def __repr__(self):
        if not hasattr(self, "scaffold"):
            return f"'{self.name}'"
        pre = [ct.name for ct in self.presynaptic.cell_types]
        post = [ct.name for ct in self.postsynaptic.cell_types]
        return f"'{self.name}', connecting {pre} to {post}"

    @abc.abstractmethod
    def connect(self, presyn_collection, postsyn_collection):
        """
        Central method of each connection strategy. Given a pair of ``HemitypeCollection``
        (one for each connection side), should connect cell population using the
        scaffold's (available as ``self.scaffold``)
        :meth:`bsb.core.Scaffold.connect_cells` method.

        :param bsb.connectivity.strategy.HemitypeCollection presyn_collection: presynaptic
            filtered cell population.
        :param bsb.connectivity.strategy.HemitypeCollection postsyn_collection:
            postsynaptic filtered cell population.
        """
        pass

    def get_deps(self):
        return set(self.depends_on)

    def _get_connect_args_from_job(self, pre_roi, post_roi):
        pre = HemitypeCollection(self.presynaptic, pre_roi)
        post = HemitypeCollection(self.postsynaptic, post_roi)
        return pre, post

    def connect_cells(self, pre_set, post_set, src_locs, dest_locs, tag=None):
        """
        Connect cells from a presynaptic placement set to cells of a postsynaptic
        placement set, and produce a unique name to describe their connectivity set. The
        description of the hemitype (source or target cell population) `connection
        location` is stored as a list of 3 ids: the cell index (in the placement set),
        morphology branch index, and the morphology branch section index. If no morphology
        is attached to the hemitype, then the morphology indexes can be set to -1.

        :param bsb.storage.interfaces.PlacementSet pre_set: presynaptic placement set
        :param bsb.storage.interfaces.PlacementSet post_set: postsynaptic placement set
        :param list[list[int, int, int]] src_locs: list of the presynaptic
            `connection location`.
        :param list[list[int, int, int]] dest_locs: list of the postsynaptic
            `connection location`.
        """
        names = self.get_output_names(pre_set.cell_type, post_set.cell_type)
        between_msg = f"between {pre_set.cell_type.name} and {post_set.cell_type.name}"
        if len(names) == 0:
            raise ConnectivityError(
                f"Connections {between_msg} have been disabled by output naming."
            )
        elif len(names) == 1:
            name = names[0]
            if tag is not None and tag != name:
                raise ConnectivityError(
                    f"Tag ('{tag}') and output name ('{name}') mismatch."
                )
        else:
            names_msg = f"{between_msg} (names: {', '.join(names)})."
            if tag is None:
                raise ConnectivityError(
                    f"No tag was given to decide between the available "
                    f"output names {names_msg}"
                )
            elif tag not in names:
                raise ConnectivityError(
                    f"Tag '{tag}' is not a valid output name {names_msg}"
                )
            else:
                name = tag

        self.scaffold.connect_cells(pre_set, post_set, src_locs, dest_locs, name)

    def get_region_of_interest(self, chunk):
        """
        Returns the list of chunks containing the potential postsynaptic neurons, based on
        a chunk containing the presynaptic neurons.

        :param chunk: Presynaptic chunk
        :type chunk: bsb.storage._chunks.Chunk
        :returns: List of postsynaptic chunks
        :rtype: list[bsb.storage._chunks.Chunk]
        """
        pass

    def queue(self, pool: JobPool):
        """
        Specifies how to queue this connectivity strategy into a job pool.

        Can be overridden, the default implementation asks each partition to chunk itself
        and creates 1 placement job per chunk.
        """
        # Get the queued jobs of all the strategies we depend on.
        dep_jobs = set(
            chain.from_iterable(
                pool.get_submissions_of(strat) for strat in self.get_deps()
            )
        )
        pre_types = self.presynaptic.cell_types
        # Iterate over each chunk that is populated by our presynaptic cell types.
        from_chunks = set(
            chain.from_iterable(
                ct.get_placement_set().get_all_chunks() for ct in pre_types
            )
        )
        rois = {
            chunk: roi
            for chunk in from_chunks
            if (roi := self.get_region_of_interest(chunk)) is None or len(roi)
        }
        if not rois:
            warn(
                f"No overlap found between {[pre.name for pre in pre_types]} and "
                f"{[post.name for post in self.postsynaptic.cell_types]} "
                f"in '{self.name}'."
            )
        for chunk, roi in rois.items():
            pool.queue_connectivity(self, [chunk], roi, deps=dep_jobs)

    def get_cell_types(self):
        return set(self.presynaptic.cell_types) | set(self.postsynaptic.cell_types)

    def get_all_pre_chunks(self):
        all_ps = (ct.get_placement_set() for ct in self.presynaptic.cell_types)
        chunks = set(ichain(ps.get_all_chunks() for ps in all_ps))
        return list(chunks)

    def get_all_post_chunks(self):
        all_ps = (ct.get_placement_set() for ct in self.postsynaptic.cell_types)
        chunks = set(ichain(ps.get_all_chunks() for ps in all_ps))
        return list(chunks)

    def get_output_names(self, pre=None, post=None):
        if (pre is None) != (post is None):
            raise RuntimeError("pre and post must be specified or omitted together.")
        if pre is not None and (
            pre not in self.presynaptic.cell_types
            or post not in self.postsynaptic.cell_types
        ):
            raise ValueError(
                f"'{pre.name}' and '{post.name}' are not a valid cell pair type "
                f"for this connectivity strategy."
            )
        if self.output_naming is None or isinstance(self.output_naming, str):
            return self._infer_output_name(self.output_naming or self.name, pre, post)
        elif isinstance(self.output_naming, list):
            # Call `_infer_output_name` for each given `base` in the list,
            # and chain them together
            return [
                *ichain(
                    self._infer_output_name(base, pre, post)
                    for base in self.output_naming
                )
            ]
        else:
            return self._get_output_name(pre, post)

    def _infer_output_name(self, base, pre, post):
        if len(self.presynaptic.cell_types) > 1 or len(self.postsynaptic.cell_types) > 1:
            if pre is None:
                # All output names
                return [
                    *ichain(
                        self._infer_output_name(base, pre_ct, post_ct)
                        for pre_ct in self.presynaptic.cell_types
                        for post_ct in self.postsynaptic.cell_types
                    )
                ]
            else:
                # Pair specific output name
                return [f"{base}_{pre.name}_to_{post.name}"]
        else:
            # Single output name
            return [base]

    def _get_output_name(self, pre, post):
        if pre is None:
            # All output names
            return [
                *ichain(
                    self._get_output_name(pre_ct, post_ct)
                    for pre_ct in self.presynaptic.cell_types
                    for post_ct in self.postsynaptic.cell_types
                )
            ]
        else:
            # Pair specific output name
            MISSING = type("MISSING", (), {"get": lambda *args: MISSING})()
            spec = self.output_naming.get(pre.name, MISSING).get(post.name, MISSING)
            if spec is MISSING:
                return self._infer_output_name(self.name, pre, post)
            elif spec is None:
                return []
            elif isinstance(spec, str):
                return [spec]
            else:
                return spec


__all__ = ["ConnectionStrategy", "Hemitype", "HemitypeCollection"]



from dataclasses import dataclass as _dc

from .adapter import SimulatorAdapter as _SimAdapter
from .simulation import Simulation as _Sim


@_dc
class SimulationBackendPlugin:
    Adapter: _SimAdapter
    Simulation: _Sim


def get_simulation_adapter(name: str, comm=None):
    """
    Return the adapter corresponding to the given simulator name.

    :param str name: Name of the simulator.
    :param comm: The mpi4py MPI communicator to use. Only nodes in the communicator will
        participate in the simulation. The first node will idle as the main node.
    """
    from ._backends import get_simulation_adapters

    return get_simulation_adapters()[name](comm=comm)


__all__ = ["SimulationBackendPlugin", "get_simulation_adapter"]



import itertools

import numpy as np

from ._util import obj_str_insert


class _lset(set):
    def __hash__(self):
        return int.from_bytes(":|\\!#è".join(sorted(self)).encode(), "little")

    def __eq__(self, other):
        return hash(self) == hash(_lset(other))

    def copy(self):
        return self.__class__(self)


class EncodedLabels(np.ndarray):
    def __new__(subtype, *args, labels=None, **kwargs):
        kwargs["dtype"] = int
        array = super().__new__(subtype, *args, **kwargs)
        if labels is None:
            labels = {0: _lset()}
        array.labels = {int(k): _lset(v) for k, v in labels.items()}
        return array

    @obj_str_insert
    def __repr__(self):
        labellist = ", ".join(
            (
                f"{sum(self == k)} labelled {list(ls)}"
                if len(ls)
                else f"{sum(self == k)} unlabelled"
            )
            for k, ls in self.labels.items()
        )
        return f"with {len(self)} elements, of which {labellist}"

    __str__ = __repr__

    def __array_finalize__(self, array):
        if array is not None:
            self.labels = getattr(array, "labels", {0: _lset()})

    def __eq__(self, other):
        try:
            return np.allclose(*EncodedLabels._merged_translate((self, other)))
        except Exception:
            return np.array(self, copy=False) == other

    @property
    def raw(self):
        return np.array(self, copy=False)

    def copy(self, *args, **kwargs):
        cp = super().copy(*args, **kwargs)
        cp.labels = {k: v.copy() for k, v in cp.labels.items()}
        return cp

    def label(self, labels, points):
        if not len(points):
            return
        _transitions = {}
        # A counter that skips existing values.
        counter = (c for c in itertools.count() if c not in self.labels)

        # This local function looks up the new id that a point should transition
        # to when `labels` are added to the labels it already has.
        def transition(point):
            nonlocal _transitions
            # Check if we already know the transition of this value.
            if point in _transitions:
                return _transitions[point]
            else:
                # First time making this transition. Join the existing and new labels
                trans_labels = self.labels[point].copy()
                trans_labels.update(labels)
                # Check if this new combination of labels already is assigned an id.
                for k, v in self.labels.items():
                    if trans_labels == v:
                        # Transition labels already exist, store and return it
                        _transitions[point] = k
                        return k
                else:
                    # Transition labels are a new combination, store them under a new id.
                    transition = next(counter)
                    self.labels[transition] = trans_labels
                    # Cache the result
                    _transitions[point] = transition
                    return transition

        # Replace the label values with the transition values
        self[points] = np.vectorize(transition)(self[points])

    def contains(self, labels):
        return np.any(self.get_mask(labels))

    def index_of(self, labels):
        for i, lset in self.labels.items():
            if lset == labels:
                return i
        else:
            raise IndexError(f"Labelset {labels} does not exist")

    def get_mask(self, labels=None):
        """
        Get indexes matching the labels provided.

        :param list[str] labels: List of labels
        :rtype: numpy.ndarray[int]
        """
        labels = set(labels or [])
        has_any = [
            k
            for k, v in self.labels.items()
            if any(lbl in v for lbl in labels) or v == labels
        ]
        return np.isin(self, has_any)

    def walk(self):
        """
        Iterate over the branch, yielding the labels of each point.
        """
        for x in self:
            yield self.labels[x].copy()

    def expand(self, label):
        """
        Translate a label value into its corresponding labelset.
        """
        return self.labels[label].copy()

    @classmethod
    def none(cls, len):
        """
        Create EncodedLabels without any labelsets.
        """
        return cls(len, buffer=np.zeros(len, dtype=int))

    @classmethod
    def from_labelset(cls, len, labelset):
        """
        Create EncodedLabels with all points labelled to the given labelset.
        """
        return cls(len, buffer=np.ones(len), labels={0: _lset(), 1: _lset(labelset)})

    @classmethod
    def concatenate(cls, *label_arrs):
        if not label_arrs:
            return EncodedLabels.none(0)
        lookups = EncodedLabels._get_merged_lookups(label_arrs)
        total = sum(len(len_) for len_ in label_arrs)
        concat = cls(total, labels=lookups[0])
        ptr = 0
        for block in EncodedLabels._merged_translate(label_arrs, lookups):
            nptr = ptr + len(block)
            # Concatenate the translated block
            concat[ptr:nptr] = block
            ptr = nptr
        return concat

    @staticmethod
    def _get_merged_lookups(arrs):
        if not arrs:
            return {0: _lset()}
        merged = {}
        new_labelsets = set()
        to_map_arrs = {}
        for arr in arrs:
            for k, l_ in arr.labels.items():
                if k not in merged:
                    # The label spot is available, so take it
                    merged[k] = l_
                elif merged[k] != l_:
                    # The labelset doesn't match, so this array will have to be mapped,
                    # and a new spot found for the conflicting labelset.
                    new_labelsets.add(l_)
                    # np ndarray unhashable, for good reason, so use `id()` for quick hash
                    to_map_arrs[id(arr)] = arr
                # else: this labelset matches with the superset's nothing to do

        # Collect new spots for new labelsets
        counter = (c for c in itertools.count() if c not in merged)
        lset_map = {}
        for labelset in new_labelsets:
            key = next(counter)
            merged[key] = labelset
            lset_map[labelset] = key

        return merged, to_map_arrs, lset_map

    def _merged_translate(arrs, lookups=None):
        if lookups is None:
            merged, to_map_arrs, lset_map = EncodedLabels._get_merged_lookups(arrs)
        else:
            merged, to_map_arrs, lset_map = lookups
        for arr in arrs:
            if id(arr) not in to_map_arrs:
                # None of the label array's labelsets need to be mapped, good as is.
                block = arr
            else:
                # Lookup each labelset, if found, map to new value, otherwise, map to
                # original value.
                arrmap = {og: lset_map.get(lset, og) for og, lset in arr.labels.items()}
                block = np.vectorize(arrmap.get)(arr)
            yield block



import numpy as np

from ... import config
from ...config import types
from .. import ConnectionStrategy
from .shape_morphology_intersection import _create_geometric_conn_arrays
from .shape_shape_intersection import ShapeHemitype


def overlap_boxes(box1_min, box1_max, box2_min, box2_max):
    """
    Check if two minimal bounding box are overlapping.

    :param numpy.ndarray box1_min: 3D point representing the lowest coordinate of the
        minimal bounding box.
    :param numpy.ndarray box1_max: 3D point representing the highest coordinate of the
        minimal bounding box.
    :param numpy.ndarray box2_min: 3D point representing the lowest coordinate of the
        minimal bounding box.
    :param numpy.ndarray box2_max: 3D point representing the highest coordinate of the
        minimal bounding box.
    """
    return np.all((box1_max >= box2_min) & (box2_max >= box1_min))


@config.node
class MorphologyToShapeIntersection(ConnectionStrategy):
    postsynaptic = config.attr(type=ShapeHemitype, required=True)
    affinity = config.attr(type=types.fraction(), required=True, hint=0.1)
    """
    Ratio of apositions to keep over the total number of contact points.
    """
    pruning_ratio = config.attr(type=types.fraction(), required=True, hint=0.1)
    """
    Ratio of conections to keep over the total number of apositions.
    """

    def get_region_of_interest(self, chunk):
        lpre, upre = self.presynaptic._get_rect_ext(tuple(chunk.dimensions))
        post_chunks = self.postsynaptic.get_all_chunks()
        tree = self.postsynaptic._get_shape_boxtree(
            post_chunks,
        )
        pre_mbb = [
            np.concatenate(
                [
                    (lpre + chunk) * chunk.dimensions,
                    np.maximum((upre + chunk), (lpre + chunk) + 1) * chunk.dimensions,
                ]
            )
        ]
        return [post_chunks[i] for i in tree.query(pre_mbb, unique=True)]

    def connect(self, pre, post):
        for pre_ps in pre.placement:
            for post_ps in post.placement:
                self._connect_type(pre_ps, post_ps)

    def _connect_type(self, pre_ps, post_ps):
        pre_pos = pre_ps.load_positions()
        post_pos = post_ps.load_positions()

        post_shapes = self.postsynaptic.shapes_composition.__copy__()

        to_connect_pre = np.empty([0, 3], dtype=int)
        to_connect_post = np.empty([0, 3], dtype=int)

        morpho_set = pre_ps.load_morphologies()
        pre_morphos = morpho_set.iter_morphologies(cache=True, hard_cache=True)

        for pre_id, (pre_coord, morpho) in enumerate(
            zip(pre_pos, pre_morphos, strict=False)
        ):
            # Get the branches
            branches = morpho.get_branches()

            # Build ids array from the morphology
            pre_points_ids, pre_morpho_coord = _create_geometric_conn_arrays(
                branches, pre_id, pre_coord
            )
            pre_min_mbb, pre_max_mbb = morpho.bounds
            pre_min_mbb += pre_coord
            pre_max_mbb += pre_coord

            tmp_pre_selection = np.full(
                [len(post_pos) * int(len(pre_morpho_coord) * self.affinity), 3],
                -1,
                dtype=int,
            )
            tmp_post_selection = np.full(
                [len(post_pos) * int(len(pre_morpho_coord) * self.affinity), 3],
                -1,
                dtype=int,
            )
            ptr = 0

            for post_id, post_coord in enumerate(post_pos):
                post_shapes.translate(post_coord)
                if overlap_boxes(
                    post_shapes.get_mbb_min(),
                    post_shapes.get_mbb_max(),
                    pre_min_mbb,
                    pre_max_mbb,
                ):
                    mbb_check = post_shapes.inside_mbox(pre_morpho_coord)
                    if np.any(mbb_check):
                        inside_pts = post_shapes.inside_shapes(
                            pre_morpho_coord[mbb_check]
                        )
                        if np.any(inside_pts):
                            local_selection = (pre_points_ids[mbb_check])[inside_pts]
                            if self.affinity < 1 and len(local_selection) > 0:
                                nb_sources = np.max(
                                    [
                                        1,
                                        int(
                                            np.floor(self.affinity * len(local_selection))
                                        ),
                                    ]
                                )
                                chosen_targets = np.random.choice(
                                    local_selection.shape[0], nb_sources
                                )
                                local_selection = local_selection[chosen_targets, :]

                            selected_count = len(local_selection)
                            if selected_count > 0:
                                tmp_pre_selection[ptr : ptr + selected_count, 0] = pre_id
                                tmp_post_selection[ptr : ptr + selected_count, 0] = (
                                    post_id
                                )
                                ptr += selected_count

                post_shapes.translate(-post_coord)

            to_connect_pre = np.vstack([to_connect_pre, tmp_pre_selection[:ptr]])
            to_connect_post = np.vstack([to_connect_post, tmp_post_selection[:ptr]])

        if self.pruning_ratio < 1 and len(to_connect_pre) > 0:
            ids_to_select = np.random.choice(
                len(to_connect_pre),
                int(np.floor(self.pruning_ratio * len(to_connect_pre))),
                replace=False,
            )
            to_connect_pre = to_connect_pre[ids_to_select]
            to_connect_post = to_connect_post[ids_to_select]

        self.connect_cells(pre_ps, post_ps, to_connect_pre, to_connect_post)


__all__ = ["MorphologyToShapeIntersection"]
